#!/usr/bin/env python

import unittest
import os
import smact
from smact.properties import compound_electroneg, band_gap_Harrison
from smact.builder import wurtzite
import smact.screening
import smact.lattice
import smact.lattice_parameters
import smact.distorter
import smact.oxidation_states
from pymatgen.core.periodic_table import Specie
from smact import Species

class TestSequenceFunctions(unittest.TestCase):

    # ---------------- TOP-LEVEL ----------------

    def test_Element_class_Pt(self):
        Pt = smact.Element('Pt')
        self.assertEqual(Pt.name, 'Platinum')
        self.assertEqual(Pt.ionpot, 8.95883)
        self.assertEqual(Pt.number, 78)
        self.assertEqual(Pt.dipol, 44.00)

    def test_ordered_elements(self):
        self.assertEqual(
            smact.ordered_elements(65, 68),
            ['Tb', 'Dy', 'Ho', 'Er'])
        self.assertEqual(
            smact.ordered_elements(52, 52),
            ['Te'])

    def test_element_dictionary(self):
        newlist = ['O', 'Rb', 'W']
        dictionary = smact.element_dictionary(newlist)
        self.assertEqual(dictionary['O'].crustal_abundance, 461000.0)
        self.assertEqual(dictionary['Rb'].oxidation_states, [-1, 1])
        self.assertEqual(dictionary['W'].name, 'Tungsten')
        self.assertTrue('Rn' in smact.element_dictionary())

    def test_are_eq(self):
        self.assertTrue(
            smact.are_eq([1.00, 2.00, 3.00],
                         [1.001, 1.999, 3.00],
                         tolerance=1e-2))
        self.assertFalse(
            smact.are_eq([1.00, 2.00, 3.00],
                         [1.001, 1.999, 3.00]))

    def test_gcd_recursive(self):
        self.assertEqual(
            smact._gcd_recursive(4, 12, 10, 32),
            2)
        self.assertEqual(
            smact._gcd_recursive(15, 4),
            1)

    def test_isneutral(self):
        self.assertTrue(
            smact._isneutral((-2, 1), (2, 4)))
        self.assertFalse(
            smact._isneutral((4, -1), (1, 3)))

    def test_neutral_ratios(self):
        ox = [1, -2, 1]
        is_neutral, neutral_combos = smact.neutral_ratios(ox)
        self.assertTrue((is_neutral))
        self.assertEqual(len(neutral_combos), 9)
        self.assertTrue((3, 2, 1) in neutral_combos)

    # ---------------- Properties ----------------

    def test_compound_eneg_brass(self):
        self.assertAlmostEqual(compound_electroneg(
            elements=["Cu", "Zn"], stoichs=[0.5, 0.5],
            source='Pauling'),
            5.0638963259)

    def test_harrison_gap_MgCl(self):
        self.assertAlmostEqual(band_gap_Harrison(
            'Mg','Cl',verbose=False,distance=2.67),
            3.545075110572662)

    # ---------------- BUILDER ----------------

    def test_builder_ZnS(self):
        ZnS, sys_ZnS = wurtzite(['Zn', 'S'])
        self.assertEqual((ZnS.sites[0].position[2]), 0)
        self.assertEqual((ZnS.sites[0].position[0]), 2./3.)

    # ---------------- SCREENING ----------------

    def test_pauling_test(self):
        Sn, S = (smact.Element(label) for label in ('Sn', 'S'))
        self.assertTrue(smact.screening.pauling_test(
            (+2, -2), (Sn.pauling_eneg, S.pauling_eneg),
            ))
        self.assertFalse(smact.screening.pauling_test(
            (-2, +2), (Sn.pauling_eneg, S.pauling_eneg)
            ))
        self.assertFalse(smact.screening.pauling_test(
            (-2, -2, +2), (S.pauling_eneg, S.pauling_eneg, Sn.pauling_eneg),
            symbols=('S', 'S', 'Sn'), repeat_anions=False
            ))
        self.assertTrue(smact.screening.pauling_test(
            (-2, -2, +2), (S.pauling_eneg, S.pauling_eneg, Sn.pauling_eneg),
            symbols=('S', 'S', 'Sn'), repeat_cations=False
            ))
        self.assertFalse(smact.screening.pauling_test(
            (-2, +2, +2), (S.pauling_eneg, Sn.pauling_eneg, Sn.pauling_eneg),
            symbols=('S', 'Sn', 'Sn'), repeat_cations=False
            ))
        self.assertTrue(smact.screening.pauling_test(
            (-2, +2, +2), (S.pauling_eneg, Sn.pauling_eneg, Sn.pauling_eneg),
            symbols=('S', 'Sn', 'Sn'), repeat_anions=False
            ))

    def test_pauling_test_old(self):
        Sn, S = (smact.Element(label) for label in ('Sn', 'S'))
        self.assertTrue(smact.screening.pauling_test_old(
            (+2, -2), (Sn.pauling_eneg, S.pauling_eneg),
            symbols=('S', 'S', 'Sn')
            ))
        self.assertFalse(smact.screening.pauling_test_old(
            (-2, +2), (Sn.pauling_eneg, S.pauling_eneg),
            symbols=('S', 'S', 'Sn')
            ))
        self.assertFalse(smact.screening.pauling_test_old(
            (-2, -2, +2), (S.pauling_eneg, S.pauling_eneg, Sn.pauling_eneg),
            symbols=('S', 'S', 'Sn'), repeat_anions=False
            ))
        self.assertTrue(smact.screening.pauling_test_old(
            (-2, -2, +2), (S.pauling_eneg, S.pauling_eneg, Sn.pauling_eneg),
            symbols=('S', 'S', 'Sn'), repeat_cations=False
            ))
        self.assertFalse(smact.screening.pauling_test_old(
            (-2, +2, +2), (S.pauling_eneg, Sn.pauling_eneg, Sn.pauling_eneg),
            symbols=('S', 'Sn', 'Sn'), repeat_cations=False
            ))
        self.assertTrue(smact.screening.pauling_test_old(
            (-2, +2, +2), (S.pauling_eneg, Sn.pauling_eneg, Sn.pauling_eneg),
            symbols=('S', 'Sn', 'Sn'), repeat_anions=False
            ))

    def test_eneg_states_test(self):
        Na, Fe, Cl = (smact.Element(label) for label in ('Na', 'Fe', 'Cl'))
        self.assertTrue(smact.screening.eneg_states_test(
            [1, 3, -1], [Na.pauling_eneg, Fe.pauling_eneg, Cl.pauling_eneg]
            ))
        self.assertFalse(smact.screening.eneg_states_test(
            [-1, 3, 1], [Na.pauling_eneg, Fe.pauling_eneg, Cl.pauling_eneg]
            ))

    def test_eneg_states_test_alternate(self):
        Na, Fe, Cl = (smact.Element(label) for label in ('Na', 'Fe', 'Cl'))
        self.assertTrue(smact.screening.eneg_states_test_alternate(
            [1, 3, -1], [Na.pauling_eneg, Fe.pauling_eneg, Cl.pauling_eneg]
            ))
        self.assertFalse(smact.screening.eneg_states_test_alternate(
            [-1, 3, 1], [Na.pauling_eneg, Fe.pauling_eneg, Cl.pauling_eneg]
            ))

    def test_eneg_states_test_threshold(self):
        self.assertFalse(smact.screening.eneg_states_test_threshold(
            [1, -1], [1.83, 1.82], threshold=0
            ))
        self.assertTrue(smact.screening.eneg_states_test_threshold(
            [1, -1], [1.83, 1.82], threshold=0.1
            ))

    def test_ml_rep_generator(self):
        Pb, O = (smact.Element(label) for label in ('Pb', 'O'))
        PbO2_ml = [0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.6666666666666666,
        0.0, 0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,
        0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,
        0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,
        0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,
        0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,
        0.3333333333333333,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,
        0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0]
        self.assertEqual(smact.screening.ml_rep_generator(
            ['Pb', 'O'], [1, 2]), PbO2_ml
            )
        self.assertEqual(smact.screening.ml_rep_generator(
            [Pb, O], [1, 2]), PbO2_ml
            )

    def test_smact_filter(self):
        Na, Fe, Cl = (smact.Element(label) for label in ('Na', 'Fe', 'Cl'))
        self.assertEqual(smact.screening.smact_filter(
            [Na, Fe, Cl], threshold=2),
            [(('Na', 'Fe', 'Cl'), (1, -1, -1), (2, 1, 1)),
            (('Na', 'Fe', 'Cl'), (1, 1, -1), (1, 1, 2))]
            )
        self.assertEqual(len(smact.screening.smact_filter(
            [Na, Fe, Cl], threshold=8)), 77
            )

    # ---------------- Lattice ----------------
    def test_Lattice_class(self):
        site_A = smact.lattice.Site([0, 0, 0], -1)
        site_B = smact.lattice.Site([0.5, 0.5, 0.5], [+2, +3])
        test_lattice = smact.lattice.Lattice(
        [site_A, site_B], space_group=221)
        self.assertEqual(test_lattice.sites[0].oxidation_states,-1)
        self.assertEqual(test_lattice.sites[1].position,[0.5,0.5,0.5])

    # ---------- Lattice parameters -----------
    def test_lattice_parameters(self):
        perovskite = smact.lattice_parameters.cubic_perovskite(
            [1.81, 1.33, 1.82])
        wurtz = smact.lattice_parameters.wurtzite(
            [1.81, 1.33])
        self.assertAlmostEqual(perovskite[0], 6.3)
        self.assertAlmostEqual(perovskite[1], 6.3)
        self.assertAlmostEqual(perovskite[3], 90)
        self.assertAlmostEqual(wurtz[0], 5.13076)
        self.assertAlmostEqual(wurtz[2], 8.3838)

    # ---------- Lattice parameters -----------
    def test_oxidation_states(self):
        ox  = smact.oxidation_states.Oxidation_state_probability_finder()
        self.assertAlmostEqual(ox.compound_probability(
        [Specie('Fe',+3), Specie('O',-2)]),
        0.74280230326)
        self.assertAlmostEqual(ox.pair_probability(
        Species('Fe',+3), Species('O',-2)),
        0.74280230326)
        self.assertEqual(len(ox.get_included_species()), 173)



if __name__ == '__main__':
    unittest.main()



"""Structure prediction implementation.

Todo:
    * Test with a fully populated database.
    * Implement n-ary substitution probabilities;
      at the moment, only zero- and single-species
      substitutions are considered.

"""

import itertools
import numpy as np
from typing import Generator, List, Tuple, Optional

from .database import StructureDB
from .mutation import CationMutator
from .structure import SmactStructure
from .utilities import parse_spec, unparse_spec


class StructurePredictor:
    """Provides structure prediction functionality.

    Implements a statistically-based model for determining
    likely structures of a given composition, based on a
    database of known compositions and a lambda table of
    weights.

    Based on the algorithm presented in:
        Hautier, G., Fischer, C., Ehrlacher, V., Jain, A., and Ceder, G. (2011)
        Data Mined Ionic Substitutions for the Discovery of New Compounds.
        Inorganic Chemistry, 50(2), 656-663.
        `doi:10.1021/ic102031h <https://pubs.acs.org/doi/10.1021/ic102031h>`_

    """

    def __init__(self, mutator: CationMutator, struct_db: StructureDB, table: str):
        """Initialize class.

        Args:
            mutator: A :class:`CationMutator` for probability calculations.
            struct_db: A :class:`StructureDB` from which to read strucutures
                to attempt to mutate.
            table: The table to reference within the database

        """
        self.cm = mutator
        self.db = struct_db
        self.table = table

    def predict_structs(
      self,
      species: List[Tuple[str, int]],
      thresh: Optional[float] = 1e-3,
      include_same: Optional[bool] = True,
    ) -> Generator[Tuple[SmactStructure, float, SmactStructure], None, None]:
        """Predict structures for a combination of species.

        Args:
            species: A list of (element, charge). The constituent species
                of the target compound.
            thresh: The probability threshold, below which to discard
                predictions.
            include_same: Whether to include unmodified structures
                from the database, i.e. structures containing all the
                same species. Defaults to True.

        Yields:
            Potential structures, as tuples of (structure, probability, parent).

        """
        # For now, consider just structures with the same species, and unary substitutions.
        # This means we need only consider structures with a difference of 0 or 1 species.

        if include_same:
            for identical in self.db.get_with_species(species, self.table):
                yield (identical, 1.0, identical)

        sub_spec = itertools.combinations(species, len(species) - 1)
        sub_spec = list(map(list, sub_spec))

        potential_unary_parents: List[List[SmactStructure]] = list(
          self.db.get_with_species(specs, self.table) for specs in sub_spec
        )

        for spec_idx, parents in enumerate(potential_unary_parents):
            # Get missing ion
            # Ensure a different ion is obtained
            if len(set(species) - set(sub_spec[spec_idx])) <1:
               continue
            (diff_spec, ) = set(species) - set(sub_spec[spec_idx])
            diff_spec_str = unparse_spec(diff_spec)

            # Determine conditional substitution likelihoods
            diff_sub_probs = self.cm.cond_sub_probs(diff_spec_str)

            for parent in parents:
                # print("Testing parent")
                # Filter out any structures with identical species
                if parent.has_species(diff_spec):
                    continue

                # Ensure parent has as many species as target
                if len(parent.species) != len(species):
                    continue

                ## Determine probability
                # Get species to be substituted
                # Ensure only 1 species is obtained
                if len(set(parent.get_spec_strs()) - set(map(unparse_spec, species)) - {diff_spec_str})>1:
                     continue
                (alt_spec, ) = (
                  set(parent.get_spec_strs()) - set(map(unparse_spec, species)) - {diff_spec_str}
                )

                if parse_spec(alt_spec)[1] != diff_spec[1]:
                    # Different charge
                    continue

                try:
                    p = diff_sub_probs.loc[alt_spec]
                except:
                    # Not in the Series
                    continue

                if p > thresh:
                    try:
                        mutated = self.cm._mutate_structure(parent, alt_spec, diff_spec_str)
                    except ValueError:
                        # Poorly decorated
                        continue
                    yield (
                      self.cm._mutate_structure(parent, alt_spec, diff_spec_str),
                      p,
                      parent, )


    def nary_predict_structs(
        self,
        species: List[Tuple[str, int]],
        n_ary: Optional[int] = 2,
        thresh: Optional[float] = 1e-3,
        include_same: Optional[bool] = True,
    ) -> Generator[Tuple[SmactStructure, float, SmactStructure], None, None]:
        """Predicts structures for a combination of species.
        
        Args:
            species: A list of (element, charge). The constituent species
             of the target compound.
            thresh: The probability threshold, below which to discard predictions.
            n_ary: The number of species in a parent compound to replace.
            include_same: Whether to include unmodified structures from the database,
             i.e. structures containing all the same species.
             
        Yields:
            Potential structures, as tuples of (structure, probability, parent).
            """
        
        if include_same:
            for identical in self.db.get_with_species(species, self.table):
                yield (identical, 1.0, identical)
        
        #Ensure that we can obtain a subset of species of the target compound
        if len(species) - n_ary ==0:
            return None
        sub_species = itertools.combinations(species, len(species)- n_ary)
        sub_species = list(map(list, sub_species))

        potential_nary_parents: List[List[SmactStructure]] = list(
            self.db.get_with_species(specs, self.table) for specs in sub_species
        )
        
        for spec_idx, parents in enumerate(potential_nary_parents):
            #Get missing ions
            #Ensure we get the correct number of species
             #Ensure the ions obtained are different
            #if len(set(species) - set(sub_species[spec_idx])) !=2:
                #continue
            diff_species = list(set(species) - set(sub_species[spec_idx]))
            diff_spec_str=[unparse_spec(i) for i in diff_species]
    
            diff_sub_probs=[self.cm.cond_sub_probs(i) for i in diff_spec_str]
    
        for parent in parents:
            #print("testing parent")
            # Filter out any structures with identical species
            if n_ary==1:
                if parent.has_species(diff_species[0]):
                    continue
            elif n_ary==2:
                if parent.has_species(diff_species[0]) and parent.has_species(diff_species[1]):
                    continue
            elif n_ary==3:
                if parent.has_species(diff_species[0]) and parent.has_species(diff_species[1]) and parent.has_species(diff_species[2]):
                    continue
        
            #Ensure parent has as many species as target
            if len(parent.species) != len(species):
                continue
        
            ## Determine probability
            # Get species to be substituted
            #Ensure n species are obtained
        
            if len(set(parent.get_spec_strs()) - set(map(unparse_spec, species)) - set(diff_species))!=n_ary:
                continue
            alt_spec=list(set(parent.get_spec_strs()) - set(map(unparse_spec, species)) -set(diff_species))
        
            #Need to consider p(A,X)p(B,Y) and p(A,Y)p(B,X)
            #if utilities.parse_spec(alt_spec_1)[1] != diff_species_1[1] and utilities.parse_spec(alt_spec_2)[1] != diff_species_2[1] :
                #Different charge
                #continue
        
            try:
                p=[]
                for i in range(n_ary):
                    p.append(diff_sub_probs[i].loc[alt_spec[i]])
            except:
                #Not in the Series
                continue

            p=np.prod(p)
            
        
            if p>thresh:
                try:
                    mutated=self.cm._nary_mutate_structure(parent,alt_spec, diff_spec_str)
                
                except ValueError:
                    #Poorly decorated
                    continue
                yield (
                    self.cm._nary_mutate_structure(parent, alt_spec, diff_spec_str),
                    p,
                    parent,
                )
            
        


#!/usr/bin/env python
# Using the ase spacegroup module this can build the structure, from
# the composition, as defined in the smact_lattice module.
# TODO:
# Estimate the cell parameters based on radii from tables.
# Add further types, Spinnel, Flourite, Delafossite ....

# Implement Structure class, c.f. dev_docs.

from ase.spacegroup import crystal
from smact.lattice import Lattice, Site


def cubic_perovskite(species, cell_par=[6, 6, 6, 90, 90, 90], repetitions=[1, 1, 1]):
    """
    Build a perovskite cell using the crystal function in ASE.

    Args:
        species (str): Element symbols
        cell_par (list): Six floats/ints specifying 3 unit cell lengths and 3 unit cell angles.
        repetitions (list): Three floats specifying the expansion of the cell in x,y,z directions.
    Returns:
        SMACT Lattice object of the unit cell,
        ASE crystal system of the unit cell.

    """
    system = crystal((species),
                     basis=[(0, 0, 0), (0.5, 0.5, 0.5), (0.5, 0.5, 0)],
                     spacegroup=221,
                     size=repetitions,
                     cellpar=cell_par)

    sites_list = []
    oxidation_states = [[2]] + [[4]] + [[-2]] * 3
    for site in zip(system.get_scaled_positions(), oxidation_states):
        sites_list.append(Site(site[0], site[1]))

    return Lattice(sites_list, oxidation_states), system


def wurtzite(species, cell_par=[2, 2, 6, 90, 90, 120], repetitions=[1, 1, 1]):
    """
    Build a wurzite cell using the crystal function in ASE.

    Args:
        species (str): Element symbols
        cell_par (list): Six floats/ints specifying 3 unit cell lengths and 3 unit cell angles.
        repetitions (list): Three floats specifying the expansion of the cell in x,y,z directions.
    Returns:
        SMACT Lattice object of the unit cell,
        ASE crystal system of the unit cell.

    """
    system = crystal((species),
                     basis=[(2. / 3., 1. / 3., 0), (2. / 3., 1. / 3., 5. / 8.)],
                     spacegroup=186,
                     size=repetitions,
                     cellpar=[3, 3, 6, 90, 90, 120])

    sites_list = []
    oxidation_states = [[1], [2], [3], [4]] + [[-1], [-2], [-3], [-4]]

    for site in zip(system.get_scaled_positions(), oxidation_states):
        sites_list.append(Site(site[0], site[1]))
    return Lattice(sites_list, oxidation_states), system



#!/usr/bin/env python
from builtins import object
import numpy as np
import smact

class Lattice(object):
      """A unique set of Sites.

      Lattice objects define a general crystal structure, with a space group and
      a collection of Site objects. These Site objects have their own fractional
      coordinates and a list of possible oxidation states (see the Site class).

      Specific crystal structures with elements assigned to sites are
      "materials" and use the Atoms class from the Atomic Simulation
      Environment.

      Attributes:
          basis_sites: A list of Site objects [SiteA, SiteB, SiteC, ...]
          comprising the basis sites in Cartesian coordinates

          space_group: Integer space group number according to the
          International Tables for Crystallography.

          structurbericht:
          Structurbericht identity, if applicable (e.g. 'B1')

      Methods:
          lattice_vector_calc():

      """

      def __init__(self, sites, space_group=1, strukturbericht=False):
            self.sites = sites
            self.space_group = space_group
            self.strukturbericht = strukturbericht

class Site(object):
      """
      A single lattice site with a list of possible oxidation states.

      The Site object is primarily used within Lattice objects.

      Attributes:
          position: A list of fractional coordinates [x,y,z]
          oxidation_states: A list of possible oxidation states e.g. [-1,0,1]

      """

      def __init__(self, position, oxidation_states=[0]):
            self.position = position
            self.oxidation_states = oxidation_states




"""Miscellaneous tools for data parsing."""

import re
from typing import Tuple


def parse_spec(species: str) -> Tuple[str, int]:
    """Parse a species string into its element and charge.

    Args:
        species (str): String representation of a species in
            the format {element}{absolute_charge}{sign}.

    Returns:
        A tuple of (element, signed_charge).

    Examples:
        >>> parse_spec("Fe2+")
        ('Fe', 2)
        >>> parse_spec("O2-")
        ('O', -2)

    """
    ele = re.match(r"[A-Za-z]+", species).group(0)

    charge_match = re.search(r"\d+", species)
    charge = int(charge_match.group(0)) if charge_match else 0

    if '-' in species:
        charge *= -1

    return ele, charge


def unparse_spec(species: Tuple[str, int]) -> str:
    """Unparse a species into a string representation.

    The analogue of :func:`parse_spec`.

    Args:
        A tuple of (element, signed_charge).

    Returns:
        String of {element}{absolute_charge}{sign}.

    Examples:
        >>> unparse_spec(("Fe", 2))
        'Fe2+'
        >>> unparse_spec(("O", -2))
        'O2-'

    """
    return f"{species[0]}{abs(species[1])}{get_sign(species[1])}"


def get_sign(charge: int) -> str:
    """Get string representation of a number's sign.

    Args:
        charge (int): The number whose sign to derive.

    Returns:
        sign (str): either '+', '-', or '' for neutral.

    """
    if charge > 0:
        return '+'
    elif charge < 0:
        return '-'
    else:
        return ''




"""
smact.distorter: Module for generating symmetry-unique substitutions on a given sub-lattice.

As input it takes the ASE crystal object (as built by smact.builder)
and the sub-lattice on which substitutions are to be made.
There is an example of how to use the code in Example_distort.py

TODO: Add a functionality to check two Atoms objects against one another
for equivalence.
"""

import copy
import smact
try:
    from pyspglib import spglib
except ImportError:
    try:
        from spglib import spglib
    except ImportError:
        raise Exception("Could not load spglib")

from ase.spacegroup import Spacegroup


def get_sg(lattice):

    """
    Get the space-group of the system.

    Args:
        lattice: the ASE crystal class
    Returns:
        sg (int): integer number of the spacegroup
    """
    spacegroup = spglib.get_spacegroup(lattice, symprec=1e-5)
    space_split=spacegroup.split()
    spg_num = space_split[1].replace('(','').replace(')','')
    sg = Spacegroup(int(spg_num))
    return sg

def get_inequivalent_sites(sub_lattice, lattice):
    """Given a sub lattice, returns symmetry unique sites for substitutions.

    Args:
        sub_lattice (list of lists): array containing Cartesian coordinates
            of the sub-lattice of interest

        lattice (ASE crystal): the total lattice

    Returns:
        List of sites
    """
    sg = get_sg(lattice)
    inequivalent_sites = []
    for site in sub_lattice:
        new_site = True
# Check against the existing members of the list of inequivalent sites
        if len(inequivalent_sites) > 0:
            for inequiv_site in inequivalent_sites:
                if smact.are_eq(site, inequiv_site) == True:
                    new_site = False
# Check against symmetry related members of the list of inequivalent sites
            equiv_inequiv_sites, _ = sg.equivalent_sites(inequiv_site)
            for equiv_inequiv_site in equiv_inequiv_sites:
                if smact.are_eq(site, equiv_inequiv_site) == True:
                    new_site = False

        if new_site == True:
            inequivalent_sites.append(site)

    return inequivalent_sites


def make_substitution(lattice,site,new_species):
    """Change atomic species on lattice site to new_species.

    Args:
        lattice (ASE crystal): Input lattice
        site (list): Cartesian coordinates of the substitution site
        new_species (str): New species

    Returns:
        lattice
     """
    i = 0
    # NBNBNBNB  It is necessary to use deepcopy for objects, otherwise changes applied to a clone
    # will also apply to the parent object.
    new_lattice = copy.deepcopy(lattice)
    lattice_sites = new_lattice.get_scaled_positions()
    for lattice_site in lattice_sites:
        if smact.are_eq(lattice_site, site):
            new_lattice[i].symbol = new_species
        i = i + 1
    return new_lattice


def build_sub_lattice(lattice, symbol):
    """Generate a sub-lattice of the lattice based on equivalent atomic species.

    Args:
        lattice (ASE crystal class): Input lattice
        symbol (string): Symbol of species identifying sub-lattice

    Returns:
        list of lists:
            sub_lattice: Cartesian coordinates of the sub-lattice of symbol
    """

    sub_lattice = []
    i = 0
    atomic_labels = lattice.get_chemical_symbols()
    positions = lattice.get_scaled_positions()
    for atom in atomic_labels:
        if atom == symbol:
            sub_lattice.append(positions[i])
        i = i + 1
    return sub_lattice



'''
\mainpage

This is the mainpage of the smact package

Authors Keith T. Butler, Adam J. Jackson, Daniel Davies, Josh Evans, Tim Gauntlet, Aron Walsh.

Semiconducting Materials from Analogy and Chemical Theory (SMACT)
'''



"""Tools for database interfacing for high throughput IO."""

import itertools
from multiprocessing import Pool
from operator import itemgetter

try:
    from pathos.pools import ParallelPool
    pathos_available = True
except ImportError:
    pathos_available = False

from typing import Generator, Sequence, List, Tuple, Dict, Union, Optional

import pymatgen
from pymatgen.ext.matproj import MPRester

from . import logger
from .structure import SmactStructure
from .utilities import get_sign
import sqlite3


class StructureDB:
    """SQLite Structure Database interface.

    Acts as a context manager for database interfacing
    and wraps several useful SQLite commands within
    methods.

    Attributes:
        db: The database name.
        conn: The database connection. Only open when
            used as a context manager.
        cur: The database connection cursor. Only usable
            when class implemented as context manager.

    Examples:
        Connecting to a database in memory:

        >>> DB = StructureDB(':memory:')
        >>> with DB as c:
        ...     _ = c.execute("CREATE TABLE test (id, val)")
        ...     c.execute("SELECT * FROM test").fetchall()
        []
        >>> DB.cur.execute("SELECT * FROM test").fetchall()
        Traceback (most recent call last):
            ...
        sqlite3.ProgrammingError: Cannot operate on a closed database.

    """

    def __init__(self, db: str):
        """Set database name.

        Args:
            db (str): The name of the database. Can also be ':memory:'
                to connect to a database in RAM.

        """
        self.db = db

    def __enter__(self) -> sqlite3.Cursor:
        """Initialize database connection.

        Returns:
            An SQLite cursor for interfacing with the database.

        """
        self.conn = sqlite3.connect(self.db)
        self.cur = self.conn.cursor()

        return self.cur

    def __exit__(self, exc_type, *args):
        """Close database connection.

        Commits all changes before closing.
        Alternatively, rolls back any changes if an exception
        was raised, causing the context to be exited.

        """
        if exc_type is not None:
            self.conn.rollback()
        else:
            self.conn.commit()

        self.conn.close()

    def add_mp_icsd(
      self,
      table: str,
      mp_data: Optional[List[Dict[str, Union[pymatgen.core.Structure, str]]]] = None,
      mp_api_key: Optional[str] = None
    ) -> int:
        """Add a table populated with Materials Project-hosted ICSD structures.

        Note:
            This is very computationally expensive for large datasets
            and will not likely run on a laptop.
            If possible, download a pre-constructed database.

        Args:
            table (str): The name of the table to add.
            mp_data: The Materials Project data to parse. If this is None, data
                will be downloaded. Downloading data needs `mp_api_key` to be set.
            mp_api_key (str): A Materials Project API key. Only needed if `mp_data`
                is None.
        
        Returns:
            The number of structs added.

        """
        if mp_data is None:  #pragma: no cover
            with MPRester(mp_api_key) as m:
                data = m.query(
                  criteria={'icsd_ids.0': {
                    '$exists': True
                  }},
                  properties=['structure', 'material_id'],
                )
        else:
            data = mp_data

        self.add_table(table)

        if pathos_available:
            pool = ParallelPool()
            parse_iter = pool.uimap(parse_mprest, data)
        else:  #pragma: no cover
            parse_iter = map(parse_mprest, data)

        return self.add_structs(parse_iter, table, commit_after_each=True)

    def add_table(self, table: str):
        """Add a table to the database.

        Args:
            table: The name of the table to add

        """
        with self as c:
            c.execute(
              f"""CREATE TABLE {table}
                (composition TEXT NOT NULL, structure TEXT NOT NULL)""",
            )

    def add_struct(self, struct: SmactStructure, table: str):
        """Add a SmactStructure to a table.

        Args:
            struct: The :class:`~.SmactStructure` to add.
            table: The name of the table to add the structure to.

        """
        entry = (struct.composition(), struct.as_poscar())

        with self as c:
            c.execute(f"INSERT into {table} VALUES (?, ?)", entry)

    def add_structs(
      self,
      structs: Sequence[SmactStructure],
      table: str,
      commit_after_each: Optional[bool] = False
    ) -> int:
        """Add several SmactStructures to a table.

        Args:
            structs: Iterable of :class:`~.SmactStructure` s to add to table.
            table: The name of the table to add the structs to.
            commit_after_each (bool, optional): Whether to commit the addition
                after each structure is added.
                This is useful when adding a large number of structures over
                a long timeframe, as it ensures some structures are added,
                even if the program terminates before completion.
                Defaults to False.

        Returns:
            The number of structures added.

        """
        with self as c:
            num = 0
            for struct in structs:
                if struct is None:
                    # Handling for poorly decorated structures
                    continue
                entry = (struct.composition(), struct.as_poscar())
                c.execute(f"INSERT into {table} VALUES (?, ?)", entry)
                num += 1

                if commit_after_each:
                    self.conn.commit()

        return num

    def get_structs(self, composition: str, table: str) -> List[SmactStructure]:
        """Get SmactStructures for a given composition.

        Args:
            composition: The composition to search for.
                See :meth:`SmactStructure.composition`.
            table: The name of the table in which to search.

        Returns:
            A list of :class:`~.SmactStructure` s.

        """
        with self as c:
            c.execute(
              f"SELECT structure FROM {table} WHERE composition = ?",
              (composition, ), )
            structs = c.fetchall()
        return [SmactStructure.from_poscar(pos[0]) for pos in structs]

    def get_with_species(
      self,
      species: List[Tuple[str, int]],
      table: str, ) -> List[SmactStructure]:
        """Get SmactStructures containing given species.

        Args:
            species: A list of species as tuples, in (element, charge) format.
            table: The name of the table from which to get the species.

        Returns:
            A list of :class:`SmactStructure` s in the table that contain the species.

        """
        glob = "*".join("{}_*_{}{}" for _ in range(len(species)))
        glob = f"*{glob}*"

        species.sort(key=itemgetter(1), reverse=True)
        species.sort(key=itemgetter(0))

        # Generate a list of [element1, charge1, sign1, element2, ...]
        vals = list(
          itertools.chain.from_iterable([x[0], abs(x[1]), get_sign(x[1])] for x in species)
        )

        glob_form = glob.format(*vals)

        with self as c:
            c.execute(
              f"SELECT structure FROM {table} WHERE composition GLOB ?",
              (glob_form, ), )
            structs = c.fetchall()

        return [SmactStructure.from_poscar(pos[0]) for pos in structs]


def parse_mprest(data: Dict[str, Union[pymatgen.core.Structure, str]], ) -> SmactStructure:
    """Parse MPRester query data to generate structures.

    Args:
        data: A dictionary containing the keys 'structure' and
            'material_id', with the associated values.

    Returns:
        An oxidation-state-decorated :class:`SmactStructure`.

    """
    try:
        return SmactStructure.from_py_struct(data["structure"])
    except:
        # Couldn't decorate with oxidation states
        logger.warn(f"Couldn't decorate {data['material_id']} with oxidation states.")


"""Test structure prediction module"""

import itertools
import json
import logging
import os
import pickle
import unittest
from contextlib import contextmanager
from operator import itemgetter
from random import sample

import numpy as np
import pandas as pd
import pymatgen
from pandas.testing import assert_frame_equal, assert_series_equal
from pymatgen.analysis.structure_prediction.substitution_probability import SubstitutionProbability

import smact
from smact import Species
from smact.structure_prediction.database import StructureDB
from smact.structure_prediction.mutation import CationMutator
from smact.structure_prediction.prediction import StructurePredictor
from smact.structure_prediction.structure import SmactStructure

files_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), "files")
TEST_STRUCT = os.path.join(files_dir, "test_struct")
TEST_POSCAR = os.path.join(files_dir, "test_poscar.txt")
TEST_PY_STRUCT = os.path.join(files_dir, "pymatgen_structure.json")
TEST_LAMBDA_JSON = os.path.join(files_dir, "test_lambda_tab.json")
TEST_LAMBDA_CSV = os.path.join(files_dir, "test_lambda_tab.csv")
TEST_MP_DATA = os.path.join(files_dir, "mp_data")
TEST_PREDICTOR_DB = os.path.join(files_dir, "test_predictor.db")
TEST_PREDICTOR_TABLE = "TEST"


def generate_test_structure(comp: str) -> bool:
    """Generate a pickled test structure for comparison."""
    poscar_file = os.path.join(files_dir, f"{comp}.txt")
    s = SmactStructure.from_file(poscar_file)

    with open(TEST_STRUCT, 'wb') as f:
        pickle.dump(s, f)

    with open(TEST_POSCAR, 'w') as f:
        f.write(s.as_poscar())

    return True


@contextmanager
def ignore_warnings(logger: logging.Logger) -> int:
    """Ignore logging warnings."""
    log_lvl_buff = logger.getEffectiveLevel()
    logger.setLevel(logging.ERROR)
    yield log_lvl_buff
    logger.setLevel(log_lvl_buff)


class StructureTest(unittest.TestCase):
    """`SmactStructure` testing."""

    TEST_SPECIES = {
      "CaTiO3": [('Ca', 2, 1), ('Ti', 4, 1), ('O', -2, 3)],
      "NaCl": [('Na', 1, 1), ('Cl', -1, 1)],
      "Fe": [('Fe', 0, 1)],
    }

    def assertStructAlmostEqual(self, s1: SmactStructure, s2: SmactStructure, places: int = 7):
        """Assert that two SmactStructures are almost equal.

        Almost equality dependent on how many decimal places the site coordinates
        are equal to.
        """
        must_equal = ["species", "lattice_param"]
        for cond in must_equal:
            self.assertEqual(getattr(s1, cond), getattr(s2, cond))

        self.assertTrue(np.array_equal(s1.lattice_mat, s2.lattice_mat))
        self.assertEqual(list(s1.sites.keys()), list(s2.sites.keys()))

        for si, sj in zip(s1.sites, s2.sites):
            for c1, c2 in zip(si, sj):
                self.assertAlmostEqual(c1, c2, places=places)

    def test_as_poscar(self):
        """Test POSCAR generation."""
        for comp in self.TEST_SPECIES.keys():
            with self.subTest(comp=comp):
                comp_file = os.path.join(files_dir, f"{comp}.txt")
                with open(comp_file, "r") as f:
                    struct = SmactStructure.from_file(comp_file)
                    self.assertEqual(struct.as_poscar(), f.read())

    @staticmethod
    def _gen_empty_structure(species):
        """Generate an empty set of arguments for `SmactStructure` testing."""
        lattice_mat = np.array([[0] * 3] * 3)

        if isinstance(species[0][0], str):
            species_strs = [
              "{ele}{charge}{sign}".format(
                ele=spec[0],
                charge=abs(spec[1]),
                sign='+' if spec[1] >= 0 else '-', ) for spec in species
            ]
        else:
            species_strs = [
              "{ele}{charge}{sign}".format(
                ele=spec[0].symbol,
                charge=abs(spec[0].oxidation),
                sign='+' if spec[0].oxidation >= 0 else '-',
              ) for spec in species
            ]

        sites = {spec: [[]] for spec in species_strs}
        return species, lattice_mat, sites

    def test_from_py_struct(self):
        """Test generation of SmactStructure from a pymatgen Structure."""
        with open(TEST_PY_STRUCT, 'r') as f:
            d = json.load(f)
            py_structure = pymatgen.core.Structure.from_dict(d)

        with ignore_warnings(smact.structure_prediction.logger):
            s1 = SmactStructure.from_py_struct(py_structure)

        s2 = SmactStructure.from_file(os.path.join(files_dir, "CaTiO3.txt"))

        self.assertStructAlmostEqual(s1, s2)

    def test_has_species(self):
        """Test determining whether a species is in a `SmactStructure`."""
        s1 = SmactStructure(*self._gen_empty_structure([('Ba', 2, 2), ('O', -2, 1), ('F', -1, 2)]))

        self.assertTrue(s1.has_species(('Ba', 2)))
        self.assertFalse(s1.has_species(('Ba', 3)))
        self.assertFalse(s1.has_species(('Ca', 2)))

    def test_smactStruc_comp_key(self):
        """Test generation of a composition key for `SmactStructure`s."""
        s1 = SmactStructure(*self._gen_empty_structure([('Ba', 2, 2), ('O', -2, 1), ('F', -1, 2)]))
        s2 = SmactStructure(*self._gen_empty_structure([('Fe', 2, 1), ('Fe', 3, 2), ('O', -2, 4)]))

        Ba = Species('Ba', 2)
        O = Species('O', -2)
        F = Species('F', -1)
        Fe2 = Species('Fe', 2)
        Fe3 = Species('Fe', 3)

        s3 = SmactStructure(*self._gen_empty_structure([(Ba, 2), (O, 1), (F, 2)]))
        s4 = SmactStructure(*self._gen_empty_structure([(Fe2, 1), (Fe3, 2), (O, 4)]))

        Ba_2OF_2 = "Ba_2_2+F_2_1-O_1_2-"
        Fe_3O_4 = "Fe_2_3+Fe_1_2+O_4_2-"
        self.assertEqual(s1.composition(), Ba_2OF_2)
        self.assertEqual(s2.composition(), Fe_3O_4)
        self.assertEqual(s3.composition(), Ba_2OF_2)
        self.assertEqual(s4.composition(), Fe_3O_4)

    def test_smactStruc_from_file(self):
        """Test the `from_file` method of `SmactStructure`."""
        with open(TEST_STRUCT, 'rb') as f:
            s1 = pickle.load(f)

        s2 = SmactStructure.from_file(TEST_POSCAR)

        self.assertEqual(s1, s2)

    def test_equality(self):
        """Test equality determination of `SmactStructure`."""
        struct_files = [os.path.join(files_dir, f"{x}.txt") for x in ["CaTiO3", "NaCl"]]
        CaTiO3 = SmactStructure.from_file(struct_files[0])
        NaCl = SmactStructure.from_file(struct_files[1])

        with self.subTest(msg="Testing equality of same object."):
            self.assertEqual(CaTiO3, CaTiO3)

        with self.subTest(msg="Testing inequality of different types."):
            self.assertNotEqual(CaTiO3, "CaTiO3")

        with self.subTest(msg="Testing inequality of different objects."):
            self.assertNotEqual(CaTiO3, NaCl)

    def test_ele_stoics(self):
        """Test acquiring element stoichiometries."""
        s1 = SmactStructure(*self._gen_empty_structure([('Fe', 2, 1), ('Fe', 3, 2), ('O', -2, 4)]))
        s1_stoics = {'Fe': 3, 'O': 4}
        s2 = SmactStructure(*self._gen_empty_structure([('Ba', 2, 2), ('O', -2, 1), ('F', -1, 2)]))
        s2_stoics = {'Ba': 2, 'O': 1, 'F': 2}

        for test, expected in [(s1, s1_stoics), (s2, s2_stoics)]:
            with self.subTest(species=test.species):
                self.assertEqual(SmactStructure._get_ele_stoics(test.species), expected)

    @unittest.skipUnless(os.environ.get("MPI_KEY"), "requires MPI key to be set.")
    def test_from_mp(self):
        """Test downloading structures from materialsproject.org."""
        # TODO Needs ensuring that the structure query gets the same
        # structure as we have downloaded.
        api_key = os.environ.get("MPI_KEY")

        for comp, species in self.TEST_SPECIES.items():
            with self.subTest(comp=comp):
                comp_file = os.path.join(files_dir, f"{comp}.txt")
                local_struct = SmactStructure.from_file(comp_file)
                mp_struct = SmactStructure.from_mp(species, api_key)
                self.assertEqual(local_struct, mp_struct)


class StructureDBTest(unittest.TestCase):
    """Test StructureDB interface."""

    TEST_DB = os.path.join(files_dir, "test_db.tmp")
    TEST_TABLE = "Structures"
    TEST_MP_TABLE = "Structures1"

    @classmethod
    def tearDownClass(cls):
        """Remove database files."""
        if os.path.exists(cls.TEST_DB):
            os.remove(cls.TEST_DB)

    def test_db_interface(self):
        """Test interfacing with database."""
        with self.subTest(msg="Instantiating database."):
            self.db = StructureDB(self.TEST_DB)

        with self.subTest(msg="Adding table."):
            try:
                self.db.add_table(self.TEST_TABLE)
            except Exception as e:
                self.fail(e)

        struct_file = os.path.join(files_dir, "CaTiO3.txt")
        struct = SmactStructure.from_file(struct_file)

        with self.subTest(msg="Adding structure to table."):
            try:
                self.db.add_struct(struct, self.TEST_TABLE)
            except Exception as e:
                self.fail(e)

        with self.subTest(msg="Getting structure from table."):
            struct_list = self.db.get_structs(struct.composition(), self.TEST_TABLE)
            self.assertEqual(len(struct_list), 1)
            self.assertEqual(struct_list[0], struct)

        struct_files = [os.path.join(files_dir, f"{x}.txt") for x in ["NaCl", "Fe"]]
        structs = [SmactStructure.from_file(fname) for fname in struct_files]

        with self.subTest(msg="Adding multiple structures to table."):
            try:
                self.db.add_structs(structs, self.TEST_TABLE)
            except Exception as e:
                self.fail(e)

        test_with_species_args = [
          [("Na", 1)],
          [("Cl", -1)],
          [("Na", 1), ("Cl", -1)],
          [("Cl", -1), ("Na", 1)],
          [("Cl", -1)],
          [("Na", 1), ("Cl", 1)],
          [("O", -2)],
          [("Ca", 2), ("Ti", 4), ("O", -2)],
        ]

        test_with_species_exp = [
          [structs[0]],
          [structs[0]],
          [structs[0]],
          [structs[0]],
          [structs[0]],
          [],
          [struct],
          [struct],
        ]

        for spec, expected in zip(test_with_species_args, test_with_species_exp):
            with self.subTest(msg=f"Retrieving species with {spec}"):
                self.assertEqual(self.db.get_with_species(spec, self.TEST_TABLE), expected)

        with open(TEST_MP_DATA, "rb") as f:
            mp_data = pickle.load(f)

        with self.subTest(msg="Testing adding downloaded MP structures."):
            added: int = self.db.add_mp_icsd(self.TEST_MP_TABLE, mp_data)
            self.assertEqual(added, 3)


class CationMutatorTest(unittest.TestCase):
    """Test the CationMutator class."""

    @classmethod
    def setUpClass(cls):
        """Set up the test initial structure and mutator."""
        cls.test_struct = SmactStructure.from_file(TEST_POSCAR)

        cls.test_mutator = CationMutator.from_json(lambda_json=TEST_LAMBDA_JSON)
        cls.test_pymatgen_mutator = CationMutator.from_json(
          lambda_json=None, alpha=lambda x, y: -5
        )

        # 5 random test species -> 5! test pairs
        cls.test_species = sample(cls.test_pymatgen_mutator.specs, 5)
        cls.test_pairs = list(itertools.combinations_with_replacement(cls.test_species, 2))

        cls.pymatgen_sp = SubstitutionProbability(lambda_table=None, alpha=-5)

    def test_lambda_tab_pop(self):
        """Test if lambda table is populated correctly."""
        lambda_dat = [
          [-5.0, 0.5, -5.0],
          [0.5, -5.0, 0.3],
          [-5.0, 0.3, -5.0], ]
        labels = ["A", "B", "C"]

        exp_lambda = pd.DataFrame(lambda_dat, index=labels, columns=labels)

        assert_frame_equal(
          self.test_mutator.lambda_tab,
          exp_lambda,
          check_names=False, )

    def test_partition_func_Z(self):
        """Test the partition function for the whole table."""
        # 2e^0.5 + 2e^0.3 + 5e^{-5} \approx 6.0308499
        self.assertAlmostEqual(self.test_mutator.Z, 6.0308499)

    def test_pymatgen_lambda_import(self):
        """Test importing pymatgen lambda table."""
        self.assertIsInstance(self.test_pymatgen_mutator.lambda_tab, pd.DataFrame)

    def test_lambda_interface(self):
        """Test getting lambda values."""
        test_cases = [itertools.permutations(x) for x in [("A", "B"), ("A", "C"), ("B", "C")]]

        expected = [0.5, -5.0, 0.3]

        for test_case, expectation in zip(test_cases, expected):
            for spec_comb in test_case:
                s1, s2 = spec_comb
                with self.subTest(s1=s1, s2=s2):
                    self.assertEqual(self.test_mutator.get_lambda(s1, s2), expectation)

    def test_ion_mutation(self):
        """Test mutating an ion of a SmactStructure."""
        ca_file = os.path.join(files_dir, "CaTiO3.txt")
        ba_file = os.path.join(files_dir, "BaTiO3.txt")

        CaTiO3 = SmactStructure.from_file(ca_file)
        BaTiO3 = SmactStructure.from_file(ba_file)

        with self.subTest(s1="CaTiO3", s2="BaTiO3"):
            mutation = self.test_mutator._mutate_structure(CaTiO3, "Ca2+", "Ba2+")
            self.assertEqual(mutation, BaTiO3)

        na_file = os.path.join(files_dir, "NaCl.txt")
        NaCl = SmactStructure.from_file(na_file)

        with self.subTest(s1="Na1+Cl1-", s2="Na2+Cl1-"):
            with self.assertRaises(ValueError):
                self.test_mutator._mutate_structure(NaCl, "Na1+", "Na2+")

        # TODO Confirm functionality with more complex substitutions

    def test_sub_prob(self):
        """Test determining substitution probabilities."""
        for s1, s2 in self.test_pairs:
            with self.subTest(s1=s1, s2=s2):
                self.assertAlmostEqual(
                  self.pymatgen_sp.prob(s1, s2),
                  self.test_pymatgen_mutator.sub_prob(s1, s2), )

    def test_cond_sub_probs(self):
        """Test determining conditional substitution probabilities for a row."""
        for s1 in ["A", "B", "C"]:
            with self.subTest(s=s1):
                cond_sub_probs_test = self.test_mutator.cond_sub_probs(s1)

                vals = [
                  (s1, s2, self.test_mutator.cond_sub_prob(s1, s2)) for s2 in ["A", "B", "C"]
                ]

                test_df = pd.DataFrame(vals)
                test_df: pd.DataFrame = test_df.pivot(index=0, columns=1, values=2)

                # Slice to convert to series
                assert_series_equal(cond_sub_probs_test, test_df.iloc[0])

    def test_cond_sub_prob(self):
        """Test determining conditional substitution probabilities."""
        for s1, s2 in self.test_pairs:
            with self.subTest(s1=s1, s2=s2):
                self.assertAlmostEqual(
                  self.pymatgen_sp.cond_prob(s1, s2),
                  self.test_pymatgen_mutator.cond_sub_prob(s1, s2),
                )

    def test_pair_corr(self):
        """Test determining conditional substitution probabilities."""
        for s1, s2 in self.test_pairs:
            with self.subTest(s1=s1, s2=s2):
                self.assertAlmostEqual(
                  self.pymatgen_sp.pair_corr(s1, s2),
                  self.test_pymatgen_mutator.pair_corr(s1, s2),
                )

    def test_from_df(self):
        """Test creating a CationMutator from an existing DataFrame."""
        lambda_df = pd.read_csv(TEST_LAMBDA_CSV, index_col=0)

        csv_test = CationMutator(lambda_df=lambda_df)
        assert_frame_equal(
          csv_test.lambda_tab,
          self.test_mutator.lambda_tab,
          check_names=False, )

    def test_complete_cond_probs(self):
        """Test getting all conditional probabilities."""
        pairs = itertools.product(["A", "B", "C"], repeat=2)

        vals = [(
          s1,
          s2,
          self.test_mutator.cond_sub_prob(s1, s2), ) for s1, s2 in pairs]

        cond_probs = pd.DataFrame(vals)
        cond_probs = cond_probs.pivot(index=0, columns=1, values=2)

        assert_frame_equal(self.test_mutator.complete_cond_probs(), cond_probs)

    def test_complete_sub_probs(self):
        """Test getting all probabilities."""
        pairs = itertools.product(["A", "B", "C"], repeat=2)

        vals = [(
          s1,
          s2,
          self.test_mutator.sub_prob(s1, s2), ) for s1, s2 in pairs]

        sub_probs = pd.DataFrame(vals)
        sub_probs = sub_probs.pivot(index=0, columns=1, values=2)

        assert_frame_equal(self.test_mutator.complete_sub_probs(), sub_probs)

    def test_complete_pair_corrs(self):
        """Test getting all pair correlations."""
        pairs = itertools.product(["A", "B", "C"], repeat=2)

        vals = [(
          s1,
          s2,
          self.test_mutator.pair_corr(s1, s2), ) for s1, s2 in pairs]

        pair_corrs = pd.DataFrame(vals)
        pair_corrs = pair_corrs.pivot(index=0, columns=1, values=2)

        assert_frame_equal(self.test_mutator.complete_pair_corrs(), pair_corrs)


class PredictorTest(unittest.TestCase):
    """Testing for the StructurePredictor wrapper."""

    @classmethod
    def setUpClass(cls):
        """Initialize the prerequisites."""
        cls.db = StructureDB(TEST_PREDICTOR_DB)
        # NOTE: This may break if the pymatgen lambda table is updated
        cls.cm = CationMutator.from_json()
        cls.table = TEST_PREDICTOR_TABLE

    def test_prediction(self):
        """Test prediction pipeline."""
        with self.subTest(msg="Testing predictor setup"):
            sp = StructurePredictor(self.cm, self.db, self.table)

        test_specs = [("Ca", 2), ("Ti", 4), ("O", -2)]

        # At time of creation, results with thresh=0 and include_same=True are as follows:
        #
        # Ca_1_2+O_4_2-Ti_2_4+: p=1.0, from Ca_1_2+O_4_2-Ti_2_4+
        # Ca_1_2+O_3_2-Ti_1_4+: p=0.0088406421433475, from Ca_1_2+O_3_2-Si_1_4+
        # Ca_1_2+O_3_2-Ti_1_4+: p=0.05502122697602804, from Ca_1_2+Mo_1_4+O_3_2-
        # Ca_1_2+O_3_2-Ti_1_4+: p=0.0445515892332118, from Ca_1_2+O_3_2-V_1_4+
        # Ca_2_2+O_4_2-Ti_1_4+: p=0.0274412135813993, from Fe_2_2+O_4_2-Ti_1_4+

        with self.subTest(msg="Acquiring predictions"):
            try:
                predictions = list(sp.predict_structs(test_specs, thresh=0.02, include_same=False))
            except Exception as e:
                self.fail(e)

        expected_comps = ["Ca_1_2+Mo_1_4+O_3_2-", "Ca_1_2+O_3_2-V_1_4+", "Fe_2_2+O_4_2-Ti_1_4+"]
        expected_probs = [0.05502122697602804, 0.0445515892332118, 0.0274412135813993]

        # Sort from highest to lowest probability
        predictions.sort(key=itemgetter(1), reverse=True)

        parents = [x[2].composition() for x in predictions]

        with self.subTest(msg="Ensuring same parents"):
            self.assertEqual(parents, expected_comps)

        probs = [x[1] for x in predictions]

        with self.subTest(msg="Ensuring similar probabilities"):
            for p1, p2 in zip(probs, expected_probs):
                with self.subTest(p1=p1, p2=p2):
                    self.assertAlmostEqual(p1, p2)


if __name__ == "__main__":
    TestLoader = unittest.TestLoader()
    StructureTests = unittest.TestSuite()
    StructureTests.addTests(TestLoader.loadTestsFromTestCase(StructureTest))
    StructureTests.addTests(TestLoader.loadTestsFromTestCase(StructureDBTest))
    StructureTests.addTests(TestLoader.loadTestsFromTestCase(CationMutatorTest))
    StructureTests.addTests(TestLoader.loadTestsFromTestCase(PredictorTest))
    runner = unittest.TextTestRunner()
    result = runner.run(StructureTests)



from builtins import zip
from itertools import combinations
from smact import Element, neutral_ratios
import itertools

def pauling_test(oxidation_states, electronegativities,
                 symbols=[], repeat_anions=True,
                 repeat_cations=True, threshold=0.):
    """ Check if a combination of ions makes chemical sense,
        (i.e. positive ions should be of lower electronegativity).

    Args:
        ox (list):  oxidation states of elements in the compound
        paul (list): the corresponding  Pauling electronegativities
            of the elements in the compound
        symbols (list) :  chemical symbols of each site
        threshold (float): a tolerance for the allowed deviation from
            the Pauling criterion
        repeat_anions : boolean, allow an anion to repeat in different
            oxidation states in the same compound
        repeat_cations : as above, but for cations

    Returns:
        bool:
            True if positive ions have lower
            electronegativity than negative ions

"""

    if repeat_anions and repeat_cations and threshold == 0.:
        return eneg_states_test(oxidation_states,
                                electronegativities)

    elif repeat_anions and repeat_cations:
        return eneg_states_test_threshold(oxidation_states,
                                          electronegativities,
                                          threshold=threshold)

    else:
        if _no_repeats(oxidation_states, symbols,
                       repeat_anions=repeat_anions,
                       repeat_cations=repeat_cations):
            if threshold == 0.:
                return eneg_states_test(oxidation_states,
                                        electronegativities)
            else:
                return eneg_states_test_threshold(oxidation_states,
                                                  electronegativities,
                                                  threshold=threshold)
        else:
            return False


def _no_repeats(oxidation_states, symbols,
                repeat_anions=False, repeat_cations=False):
    """
    Check if any anion or cation appears twice.

    Args:
        oxidation_states (list): oxidation states of species
        symbols (list): chemical symbols corresponding to oxidation
            states
        repeat_anions (bool): If True, anions may be repeated (e.g. O
            in -1 and -2 states)
        repeat_cations (bool): if True, cations may be repeated (e.g.
            Cu in +1 and +2 states)

    Returns: bool
    """
    if repeat_anions is False and repeat_cations is False:
        return (len(symbols) == len(set(symbols)))
    else:

        anions, cations = [], []
        for state, symbol in zip(oxidation_states, symbols):
            if state > 0:
                cations.append(symbol)
            else:
                anions.append(symbol)
        if not repeat_anions and len(anions) != len(set(anions)):
            return False
        elif not repeat_cations and len(cations) != len(set(cations)):
            return False
        else:
            return True


def pauling_test_old(ox, paul, symbols, repeat_anions=True,
                     repeat_cations=True, threshold=0.):
    """ Check if a combination of ions makes chemical sense,
    (i.e. positive ions should be of lower Pauling electronegativity).
    This function should give the same results as pauling_test but is
    not optimised for speed.

    Args:
        ox (list):  oxidation states of the compound
        paul (list): the corresponding  Pauling electronegativities
            of the elements in the compound
        symbols (list) :  chemical symbols of each site.
        threshold (float): a tolerance for the allowed deviation from
            the Pauling criterion
        repeat_anions : boolean, allow an anion to repeat in different
            oxidation states in the same compound.
        repeat_cations : as above, but for cations.

    Returns:
        (bool):
            True if positive ions have lower
            electronegativity than negative ions

    """
    if None in paul:
        return False
    positive = []
    negative = []
    pos_ele = []
    neg_ele = []
    for i, state in enumerate(ox):
        if state > 0:
            if repeat_cations:
                positive.append(paul[i])
            elif symbols[i] in pos_ele:     # Reject materials where the same
                                            # cation occupies two sites.
                return False
            else:
                positive.append(paul[i])
                pos_ele.append(symbols[i])

        if state < 0:
            if repeat_anions:
                negative.append(paul[i])
            elif symbols[i] in neg_ele:     # Reject materials where the same
                                            # anion occupies two sites.
                return False
            else:
                negative.append(paul[i])
                neg_ele.append(symbols[i])

    if len(positive) == 0 or len(negative) == 0:
        return False
    if max(positive) == min(negative):
        return False
    if max(positive) - min(negative) > threshold:
        return False
    else:
        return True


def eneg_states_test(ox_states, enegs):
    """
    Internal function for checking electronegativity criterion

    This implementation is fast as it 'short-circuits' as soon as it
    finds an invalid combination. However it may be that in some cases
    redundant comparisons are made. Performance is very close between
    this method and eneg_states_test_alternate.

    Args:
        ox_states (list): oxidation states corresponding to species
            in compound
        enegs (list): Electronegativities corresponding to species in
            compound

    Returns:
        bool : True if cations have higher electronegativity than
            anions, otherwise False

    """
    for ((ox1, eneg1), (ox2, eneg2)) in combinations(list(zip(ox_states, enegs)), 2):
        if (ox1 > 0) and (ox2 < 0) and (eneg1 >= eneg2):
            return False
        elif (ox1 < 0) and (ox2 > 0) and (eneg1 <= eneg2):
            return False
        elif eneg1 is None or eneg2 is None:
            return False
    else:
        return True


def eneg_states_test_threshold(ox_states, enegs, threshold=0):
    """Internal function for checking electronegativity criterion

    This implementation is fast as it 'short-circuits' as soon as it
    finds an invalid combination. However it may be that in some cases
    redundant comparisons are made. Performance is very close between
    this method and eneg_states_test_alternate.

    A 'threshold' option is added so that this constraint may be
    relaxed somewhat.

    Args:
        ox_states (list): oxidation states corresponding to species
            in compound
        enegs (list): Electronegativities corresponding to species in
            compound
        threshold (Option(float)): a tolerance for the allowed deviation from
            the Pauling criterion

    Returns:
        bool : True if cations have higher electronegativity than
            anions, otherwise False

    """
    for ((ox1, eneg1), (ox2, eneg2)) in combinations(list(zip(ox_states, enegs)), 2):
        if (ox1 > 0) and (ox2 < 0) and ((eneg1 - eneg2) > threshold):
            return False
        elif (ox1 < 0) and (ox2 > 0) and (eneg2 - eneg1) > threshold:
            return False
    else:
        return True



def eneg_states_test_alternate(ox_states, enegs):
    """Internal function for checking electronegativity criterion

    This implementation appears to be slightly slower than
    eneg_states_test, but further testing is needed.

    Args:
        ox_states (list): oxidation states corresponding to species
            in compound
        enegs (list): Electronegativities corresponding to species in
            compound

    Returns:
        bool : True if cations have higher electronegativity than
            anions, otherwise False

    """
    min_cation_eneg, max_anion_eneg = 10, 0
    for ox_state, eneg in zip(ox_states, enegs):
        if ox_state < 1:
            min_cation_eneg = min(eneg, min_cation_eneg)
        else:
            max_anion_eneg = max(eneg, max_anion_eneg)
    return min_cation_eneg > max_anion_eneg

def ml_rep_generator(composition, stoichs=None):
    """Function to take a composition of Elements and return a
    list of values between 0 and 1 that describes the composition,
    useful for machine learning.

    The list is of length 103 as there are 103 elements
    considered in total in SMACT.

    e.g. Li2O --> [0, 0, 2/3, 0, 0, 0, 0, 1/3, 0 ....]

    Inspired by the representation used by Legrain et al. DOI: 10.1021/acs.chemmater.7b00789

    Args:
        composition (list): Element objects in composition OR symbols of elements in composition
        stoichs (list): Corresponding stoichiometries in the composition

    Returns:
        norm (list): List of floats representing the composition that sum
            to one

    """
    if stoichs == None:
        stoichs = [1 for i, el in enumerate(composition)]

    ML_rep = [0 for i in range(1,103)]
    if type(composition[0]) == Element:
        for element, stoich in zip(composition, stoichs):
            ML_rep[int(element.number)-1] += stoich
    else:
        for element, stoich in zip(composition, stoichs):
            ML_rep[int(Element(element).number)-1] += stoich

    norm = [float(i)/sum(ML_rep) for i in ML_rep]
    return norm

def smact_filter(els, threshold=8, species_unique=True, oxidation_states_set='default'):
    """Function that applies the charge neutrality and electronegativity
    tests in one go for simple application in external scripts that
    wish to apply the general 'smact test'.

    Args:
        els (tuple/list): A list of smact.Element objects
        threshold (int): Threshold for stoichiometry limit, default = 8
        species_unique (bool): Whether or not to consider elements in different
        oxidation states as unique in the results.
        oxidation_states_set (string): A string to choose which set of oxidation
        states should be chosen
    Returns:
        allowed_comps (list): Allowed compositions for that chemical system
        in the form [(elements), (oxidation states), (ratios)] if species_unique=True
        or in the form [(elements), (ratios)] if species_unique=False.
    """
    
    compositions = []

    # Get symbols and electronegativities
    symbols = tuple(e.symbol for e in els)
    electronegs = [e.pauling_eneg for e in els]

    # Select the specified oxidation states set:
    oxi_set = {
        'default':[e.oxidation_states for e in els],
        'icsd':[e.oxidation_states_icsd for e in els],
        'pymatgen': [e.oxidation_states_sp for e in els]
        }
    if oxidation_states_set in oxi_set:
        ox_combos = oxi_set[oxidation_states_set]
    else:
        raise(Exception(f'{oxidation_states_set} is not valid. Enter either "default", "icsd" or "pymatgen" for oxidation_states_set.'))
    for ox_states in itertools.product(*ox_combos):
        # Test for charge balance
        cn_e, cn_r = neutral_ratios(ox_states, threshold=threshold)
        # Electronegativity test
        if cn_e:
            electroneg_OK = pauling_test(ox_states, electronegs)
            if electroneg_OK:
                for ratio in cn_r:
                    compositions.append(tuple([symbols,ox_states,ratio]))

    # Return list depending on whether we are interested in unique species combinations
    # or just unique element combinations.
    if species_unique:
        return(compositions)
    else:
        compositions = [(i[0], i[2]) for i in compositions]
        compositions = list(set(compositions))
        return compositions



"""
Semiconducting Materials from Analogy and Chemical Theory

A collection of fast screening tools from elemental data
"""

# get correct path for datafiles when called from another directory
from builtins import filter
from builtins import map
from builtins import range
from builtins import object
from os import path
module_directory = path.abspath(path.dirname(__file__))
data_directory = path.join(module_directory, 'data')
import itertools


from math import gcd
from operator import mul as multiply

from smact import data_loader

class Element(object):
    """Collection of standard elemental properties for given element.

    Data is drawn from "data/element.txt", part of the Open Babel
    package.

    Atoms with a defined oxidation state draw properties from the
    "Species" class.

    Attributes:
        Element.symbol (string) : Elemental symbol used to retrieve data

        Element.name (string) : Full name of element

        Element.number (int) : Proton number of element

        Element.pauling_eneg (float) : Pauling electronegativity (0.0 if unknown)

        Element.ionpot (float) : Ionisation potential in eV (0.0 if unknown)

        Element.e_affinity (float) : Electron affinity in eV (0.0 if unknown)

        Element.dipol (float) : Static dipole polarizability in 1.6488e-41 C m^2 / V  (0.0 if unknown)

        Element.eig (float) : Electron eigenvalue (units unknown) N.B. For Cu, Au and Ag this defaults to d-orbital

        Element.eig_s (float) : Eigenvalue of s-orbital

        Element.SSE (float) : Solid State Energy

        Element.SSEPauling (float) : SSE based on regression fit with Pauling electronegativity

        Element.oxidation_states (list) : Default list of allowed oxidation states for use in SMACT

        Element.oxidation_states_sp (list) : List of oxdation states recognised by the Pymatgen Structure Predictor

        Element.oxidation_states_icsd (list) : List of oxidation states that appear in the ICSD

        Element.coord_envs (list): The allowed coordination enviroments for the ion

        Element.covalent_radius (float) : Covalent radius of the element

        Element.mass (float) : Molar mass of the element

        Element.crustal_abundance (float) : Crustal abundance in the earths crust mg/kg taken from CRC

        Element.HHI_p (float) : Herfindahl-Hirschman Index for elemental production

        Element.HHI_r (float) : Hirfindahl-Hirschman Index for elemental reserves

    Raises:
        NameError: Element not found in element.txt
        Warning: Element not found in Eigenvalues.csv

    """
    def __init__(self, symbol):
        """Initialise Element class

        Args:
            symbol (str): Chemical element symbol (e.g. 'Fe')

        """

        dataset = data_loader.lookup_element_data(symbol, copy=False)

        if dataset == None:
            raise NameError("Elemental data for {0} not found.".format(symbol))

        # Set coordination-environment data from the Shannon-radius data.
        # As above, it is safe to use copy = False with this Get* function.

        shannon_data = data_loader.lookup_element_shannon_radius_data(symbol, copy=False)

        if shannon_data != None:
            coord_envs = [row['coordination'] for row in shannon_data]
        else:
            coord_envs = None

        HHI_scores = data_loader.lookup_element_hhis(symbol)
        if HHI_scores == None:
            HHI_scores = (None, None)

        sse_data = data_loader.lookup_element_sse_data(symbol)
        if sse_data:
            sse = sse_data['SolidStateEnergy']
        else:
            sse = None

        sse_Pauling_data = data_loader.lookup_element_sse_pauling_data(symbol)
        if sse_Pauling_data:
            sse_Pauling = sse_Pauling_data['SolidStateEnergyPauling']
        else:
            sse_Pauling = None

        for attribute, value in (
            ('coord_envs', coord_envs),
            ('covalent_radius', dataset['r_cov']),
            ('crustal_abundance', dataset['Abundance']),
            ('e_affinity', dataset['e_affinity']),
            ('eig', dataset['p_eig']),
            ('eig_s', dataset['s_eig']),
            ('HHI_p', HHI_scores[0]),
            ('HHI_r', HHI_scores[1]),
            ('ionpot', dataset['ion_pot']),
            ('mass', dataset['Mass']),
            ('name', dataset['Name']),
            ('number', dataset['Z']),
            ('oxidation_states',
             data_loader.lookup_element_oxidation_states(symbol)),
            ('oxidation_states_icsd',
             data_loader.lookup_element_oxidation_states_icsd(symbol)),
            ('oxidation_states_sp',
             data_loader.lookup_element_oxidation_states_sp(symbol)),
            ('dipol', dataset['dipol']),
            ('pauling_eneg', dataset['el_neg']),
            ('SSE', sse),
            ('SSEPauling', sse_Pauling),
            ('symbol', symbol),
            #('vdw_radius', dataset['RVdW']),
            ):
            setattr(self, attribute, value)

class Species(Element):
    """
    Class providing data for elements in a given chemical environment

    In addition to the standard properties from the periodic table
    (inherited from the  Element class), Species objects use the
    oxidation state and coordination environment to provide further
    properties.

    Attributes:
        Species.symbol: Elemental symbol used to retrieve data

        Species.name: Full name of element

        Species.oxidation: Oxidation state of species (signed integer)

        Species.coordination: Coordination number of species (integer)

        Species.pauling_eneg: Pauling electronegativity (0.0 if unknown)

        Species.ionpot: Ionisation potential in eV (0.0 if unknown)

        Species.e_affinity: Electron affinity in eV (0.0 if unknown)

        Species.eig: Electron eigenvalue (units unknown)
            N.B. For Cu, Au and Ag this defaults to d-orbital.

        Species.shannon_radius: Shannon radius of Species.

        Species.ionic_radius: Ionic radius of Species.

    Raises:
        NameError: Element not found in element.txt
        Warning: Element not found in Eigenvalues.csv

    """

    def __init__(self,symbol,oxidation,coordination=4, radii_source="shannon"):
        Element.__init__(self,symbol)

        self.oxidation = oxidation
        self.coordination = coordination

        # Get shannon radius for the oxidation state and coordination.

        self.shannon_radius = None;

        if radii_source=="shannon":

            shannon_data = data_loader.lookup_element_shannon_radius_data(symbol);
        
        elif radii_source == "extended":
            shannon_data = data_loader.lookup_element_shannon_radius_data_extendedML(symbol)

        else:
            print("Data source not recognised. Please select 'shannon' or 'extended'. ") 

        if shannon_data:
            for dataset in shannon_data:
                if dataset['charge'] == oxidation and str(coordination) == dataset['coordination'].split('_')[0]:
                    self.shannon_radius = dataset['crystal_radius'];

        # Get ionic radius
        self.ionic_radius = None;

        if shannon_data:
            for dataset in shannon_data:
                if dataset['charge'] == oxidation and str(coordination) == dataset['coordination'].split('_')[0]:
                    self.ionic_radius = dataset['ionic_radius'];

        # Get SSE_2015 (revised) for the oxidation state.

        self.SSE_2015 = None

        sse_2015_data = data_loader.lookup_element_sse2015_data(symbol);
        if sse_2015_data:
            for dataset in sse_2015_data:
                if dataset['OxidationState'] == oxidation:
                    self.SSE_2015 = dataset['SolidStateEnergy2015']
        else:
            self.SSE_2015 = None

def ordered_elements(x,y):
    """
    Return a list of element symbols, ordered by proton number in the range x -> y
    Args:
        x,y : integers
    Returns:
        list: Ordered list of element symbols
    """
    with open(path.join(data_directory,
                        'ordered_periodic.txt'), 'r') as f:
        data = f.readlines()
    elements = []
    for line in data:
        inp = line.split()
        elements.append(inp[0])

    ordered_elements = []
    for i in range(x,y+1):
        ordered_elements.append(elements[i-1])

    return ordered_elements

def element_dictionary(elements=None):
    """
    Create a dictionary of initialised smact.Element objects

    Accessing an Element from a dict is significantly faster than
    repeadedly initialising them on-demand within nested loops.

    Args:
        elements (iterable of strings) : Elements to include. If None,
            use all elements up to 103.
    Returns:
        dict: Dictionary with element symbols as keys and smact.Element
            objects as data
    """
    if elements == None:
        elements = ordered_elements(1,103)
    return {symbol: Element(symbol) for symbol in elements}


def are_eq(A,B,tolerance=1e-4):
    """Check two arrays for tolerance [1,2,3]==[1,2,3]; but [1,3,2]!=[1,2,3]
    Args:
        A, B (lists): 1-D list of values for approximate equality comparison
        tolerance: numerical precision for equality condition

    Returns:
        boolean
    """
    are_eq = True
    if len(A) != len(B):
        are_eq = False
    else:
        i = 0
        while i < len(A):
            if abs(A[i] - B[i]) > tolerance:
                are_eq = False
            i = i + 1
    return are_eq


def lattices_are_same(lattice1, lattice2, tolerance=1e-4):
    """Checks for the equivalence of two lattices

    Args:
        lattice1,lattice2 : ASE crystal class
    Returns:
        boolean
        """
    lattices_are_same = False
    i = 0
    for site1 in lattice1:
        for site2 in lattice2:
            if site1.symbol == site2.symbol:
                if are_eq(site1.position,
                          site2.position,
                          tolerance=tolerance):
                    i += 1
    if i == len(lattice1):
        lattices_are_same = True
    return lattices_are_same


def _gcd_recursive(*args):
    """
    Get the greatest common denominator among any number of ints
    """
    if len(args) == 2:
        return gcd(*args)
    else:
        return gcd(args[0], _gcd_recursive(*args[1:]))

def _isneutral(oxidations, stoichs):
    """
    Check if set of oxidation states is neutral in given stoichiometry

    Args:
        oxidations (tuple): Oxidation states of a set of oxidised elements
        stoichs (tuple): Stoichiometry values corresponding to `oxidations`
    """
    return 0 == sum(map(multiply, oxidations, stoichs))

def neutral_ratios_iter(oxidations, stoichs=False, threshold=5):
    """
    Iterator for charge-neutral stoichiometries

    Given a list of oxidation states of arbitrary length, yield ratios in which
    these form a charge-neutral compound. Stoichiometries may be provided as a
    set of legal stoichiometries per site (e.g. a known family of compounds);
    otherwise all unique ratios are tried up to a threshold coefficient.

    Args:
        oxidations : list of integers
        stoichs : stoichiometric ratios for each site (if provided)
        threshold : single threshold to go up to if stoichs are not provided

    Yields:
        tuple: ratio that gives neutrality
    """
    if not stoichs:
        stoichs = [list(range(1,threshold+1))] * len(oxidations)

    # First filter: remove combinations which have a common denominator
    # greater than 1 (i.e. Use simplest form of each set of ratios)
    # Second filter: return only charge-neutral combinations
    return filter(
        lambda x: _isneutral(oxidations, x) and _gcd_recursive(*x) == 1,
        # Generator: enumerate all combinations of stoichiometry
        itertools.product(*stoichs)
        )

def neutral_ratios(oxidations, stoichs=False, threshold=5):
    """
    Get a list of charge-neutral compounds

    Given a list of oxidation states of arbitrary length, yield ratios in which
    these form a charge-neutral compound. Stoichiometries may be provided as a
    set of legal stoichiometries per site (e.g. a known family of compounds);
    otherwise all unique ratios are tried up to a threshold coefficient.

    Given a list of oxidation states of arbitrary length it searches for
    neutral ratios in a given ratio of sites (stoichs) or up to a given
    threshold.

    Args:
        oxidations (list of ints): Oxidation state of each site
        stoichs (list of positive ints): A selection of valid stoichiometric
            ratios for each site
        threshold (int): Maximum stoichiometry coefficient; if no 'stoichs'
            argument is provided, all combinations of integer coefficients up
            to this value will be tried.

    Returns:
        (exists, allowed_ratios) (tuple):

        exists *bool*:
            True ifc any ratio exists, otherwise False

        allowed_ratios *list of tuples*:
            Ratios of atoms in given oxidation
            states which yield a charge-neutral structure
    """
    allowed_ratios = [x for x in neutral_ratios_iter(oxidations,
                                                        stoichs=stoichs,
                                                        threshold=threshold)]
    return (len(allowed_ratios) > 0, allowed_ratios)

# List of metals
metals = ['Li','Be','Na','Mg','Al','K','Ca','Sc','Ti','V','Cr','Mn','Fe','Co','Ni',
'Cu','Zn','Ga','Ge','Rb','Sr','Y','Zr','Nb','Mo','Tc','Ru','Rh','Pd','Ag','Cd','In','Sn','Sb',
'Cs','Ba','La','Ce', 'Pr','Nd','Sm','Eu','Gd','Tb','Dy','Ho','Er','Tm','Yb',
'Lu','Hf','Ta','W','Re','Os','Ir','Pt','Au','Hg','Tl','Pb','Bi','Po','Fr','Ra','Ac',
'Th','Pa','U','Np','Pu','Am','Cm','Bk','Cf','Es','Fm','Md','No']

# List of elements that can be considered 'anions'.
# Similar to the Pymatgen 'electronegative elements' but excluding H, B, C & Si.
anions = ["N", "P", "As", "Sb",
           "O", "S", "Se", "Te",
           "F", "Cl", "Br", "I"]

# List of d-block metals
d_block = ['Sc','Ti','V','Cr','Mn','Fe','Co','Ni','Cu','Zn',
           'Y','Zr','Nb','Mo','Tc','Ru','Rh','Pd','Ag','Cd',
           'La','Hf','Ta','W','Re','Os','Ir','Pt','Au','Hg']



import unittest

import smact.tests.test
from smact.tests.test_structure import CationMutatorTest, StructureDBTest, StructureTest, PredictorTest

if __name__ == "__main__":
    loader = unittest.TestLoader()
    suite = unittest.TestSuite()
    suite.addTests(loader.loadTestsFromModule(smact.tests.test))
    suite.addTests(loader.loadTestsFromTestCase(StructureTest))
    suite.addTests(loader.loadTestsFromTestCase(StructureDBTest))
    suite.addTests(loader.loadTestsFromTestCase(CationMutatorTest))
    suite.addTests(loader.loadTestsFromTestCase(PredictorTest))
    runner = unittest.TextTestRunner()
    result = runner.run(suite)



"""Tools for handling ion mutation."""

import itertools
import json
import os
import re
from copy import deepcopy
from operator import itemgetter
from typing import Callable, Generator, Optional, Tuple

import numpy as np
import pandas as pd
import pymatgen.analysis.structure_prediction as pymatgen_sp
from .utilities import parse_spec
from .structure import SmactStructure


class CationMutator:
    """Handles cation mutation of SmactStructures based on substitution probability.
    
    Based on the algorithm presented in:
        Hautier, G., Fischer, C., Ehrlacher, V., Jain, A., and Ceder, G. (2011)
        Data Mined Ionic Substitutions for the Discovery of New Compounds.
        Inorganic Chemistry, 50(2), 656-663.
        `doi:10.1021/ic102031h <https://pubs.acs.org/doi/10.1021/ic102031h>`_
    
    """

    def __init__(
      self,
      lambda_df: pd.DataFrame,
      alpha: Optional[Callable[[str, str], float]] = (lambda s1, s2: -5.0),
    ):
        """Assign attributes and get lambda table.

        Args:
            lambda_df: A pandas DataFrame, with column and index labels
                as species strings and lambda values as entries.
            alpha: A function to call to fill in missing lambda values.
                The function must take the two species' strings as
                arguments, and return a floating point lambda
                value.
                Defaults to a function that unconditionally returns
                -5.0.

        """
        self.lambda_tab = lambda_df

        self.specs = set(
          itertools.chain.from_iterable(
            set(getattr(self.lambda_tab, x)) for x in ["columns", "index"]
          )
        )

        self.alpha = alpha

        # Make sure table is fully populated
        self._populate_lambda()

        self.Z = np.exp(self.lambda_tab.to_numpy()).sum()

    @staticmethod
    def from_json(
      lambda_json: Optional[str] = None,
      alpha: Optional[Callable[[str, str], float]] = (lambda s1, s2: -5.0),
    ):
        """Create a CationMutator instance from a DataFrame.

        Args:
            lambda_json (str, optional): JSON-style representation of the
                lambda table. This is a list of entries, containing pairs
                and their associated lambda values.
                Each entry is a list of [species1, species2, lambda].
                If not supplied, defaults to the lambda table
                included with pymatgen.
            alpha: See :meth:`__init__`.

        Returns:
            A :class:`CationMutator` instance.

        """
        if lambda_json is not None:
            with open(lambda_json, 'r') as f:
                lambda_dat = json.load(f)
        else:
            # Get pymatgen lambda table
            py_sp_dir = os.path.dirname(pymatgen_sp.__file__)
            pymatgen_lambda = os.path.join(py_sp_dir, "data", "lambda.json")
            with open(pymatgen_lambda, 'r') as f:
                lambda_dat = json.load(f)

            # Get rid of 'D1+' values to reflect pymatgen
            # implementation
            lambda_dat = [x for x in lambda_dat if 'D1+' not in x]

        # Convert lambda table to pandas DataFrame
        lambda_dat = [tuple(x) for x in lambda_dat]
        lambda_df = pd.DataFrame(lambda_dat)

        lambda_df = lambda_df.pivot(index=0, columns=1, values=2)

        return CationMutator(lambda_df, alpha)

    def _populate_lambda(self):
        """Populate lambda table.

        Ensures no values are NaN and performs alpha calculations,
        such that an entry exists for every possible species
        combination in the lambda table.
        Also ensures lambda table symmetry.

        """
        pairs = itertools.combinations_with_replacement(self.specs, 2)

        def add_alpha(s1, s2):
            """Add an alpha value to the lambda table at both coordinates."""
            a = self.alpha(s1, s2)
            self.lambda_tab.at[s1, s2] = a
            self.lambda_tab.at[s2, s1] = a

        def mirror_lambda(s1, s2):
            """Mirror the lambda value at (s2, s1) into (s1, s2)."""
            self.lambda_tab.at[s1, s2] = self.lambda_tab.at[s2, s1]

        for s1, s2 in pairs:
            try:
                if np.isnan(self.lambda_tab.at[s1, s2]):
                    try:
                        if not np.isnan(self.lambda_tab.at[s2, s1]):
                            mirror_lambda(s1, s2)
                        else:
                            add_alpha(s1, s2)
                    except KeyError:
                        add_alpha(s1, s2)
                else:
                    mirror_lambda(s2, s1)
            except KeyError:
                try:
                    if np.isnan(self.lambda_tab.at[s2, s1]):
                        add_alpha(s1, s2)
                    else:
                        mirror_lambda(s1, s2)
                except KeyError:
                    add_alpha(s1, s2)

        # Ensure symmetry
        idx = self.lambda_tab.index
        self.lambda_tab = self.lambda_tab[idx]

    def get_lambda(self, s1: str, s2: str) -> float:
        """Get lambda values corresponding to a pair of species.

        Args:
            s1 (str): One of the species.
            s2 (str): The other species.

        Returns:
            lambda (float): The lambda value, if it exists
                in the table. Otherwise, the alpha value
                for the two species.

        """
        if {s1, s2} <= self.specs:
            return self.lambda_tab.at[s1, s2]

        return self.alpha(s1, s2)

    def get_lambdas(self, species: str) -> pd.Series:
        """Get all the lambda values associated with a species.

        Args:
            species (str): The species for which to get the
                lambda values.

        Returns:
            A pandas Series, with index-labelled lambda entries.

        """
        if not {species} <= self.specs:
            raise ValueError(f"{species} not in lambda table.")

        return self.lambda_tab.loc[species]

    @staticmethod
    def _mutate_structure(
      structure: SmactStructure,
      init_species: str,
      final_species: str, ) -> SmactStructure:
        """Mutate a species within a SmactStructure.

        Replaces all instances of the species within the
        structure. Every site occupied by the species
        has its label changed to the new species,
        and the list of species of the structure is
        changed to reflect the mutation.
        Stoichiometry is maintained.
        Requires the species to have the same charge.

        Note:
            Creates a deepcopy of the supplied structure,
            such that the original instance is not modified.

        Args:
            init_species (str): The species within the structure
                to mutate.
            final_species (str): The species to replace the
                initial species with.

        Returns:
            A :class:`.~SmactStructure`, with the species
                mutated.

        """
        struct_buff = deepcopy(structure)
        init_spec_tup = parse_spec(init_species)
        struct_spec_tups = list(map(itemgetter(0, 1), struct_buff.species))
        spec_loc = struct_spec_tups.index(init_spec_tup)

        final_spec_tup = parse_spec(final_species)

        # Replace species tuple
        struct_buff.species[spec_loc] = (*final_spec_tup, struct_buff.species[spec_loc][2])

        # Check for charge neutrality
        if sum(x[1] * x[2] for x in struct_buff.species) != 0:
            raise ValueError("New structure is not charge neutral.")

        # Sort species again
        struct_buff.species.sort(key=itemgetter(1), reverse=True)
        struct_buff.species.sort(key=itemgetter(0))

        # Replace sites
        struct_buff.sites[final_species] = struct_buff.sites.pop(init_species)
        # And sort
        species_strs = struct_buff._format_style("{ele}{charge}{sign}").split(" ")
        struct_buff.sites = {spec: struct_buff.sites[spec] for spec in species_strs}

        return struct_buff

    @staticmethod
    def _nary_mutate_structure(
        structure: SmactStructure,
        init_species: list,
        final_species: list, ) -> SmactStructure:
        """Perform a n-ary mutation of a SmactStructure (n>1).
        Replaces all instances of a group of species within the structure.
        Stoichiometry is maintained.
        Charge neutrality is preserved, but the species pair do not need the same charge.
        
        Args:
            init_species (list): A list of species within the structure to mutate.
            final_species (list): The list of species to replace the initial species with
            
        """
        #Determine the number of species to mutate
        n=len(init_species)

        struct_buff=deepcopy(structure)
        init_spec_tup_list=[parse_spec(i) for i in init_species]
        struct_spec_tups = list(map(itemgetter(0,1), struct_buff.species))
        spec_loc=[struct_spec_tups.index(init_spec_tup_list[i]) for i in range(n)]
    
        final_spec_tup_list=[parse_spec(i) for i in final_species]
    
        #Replace species tuple
        for i in range(n):
            struct_buff.species[spec_loc[i]] = (*final_spec_tup_list[i], struct_buff.species[spec_loc[i]][2])

        #Check for charge neutrality
        if sum(x[1]*x[2] for x in struct_buff.species) !=0:
            raise ValueError("New structure is not charge neutral")


        #Sort species again
        struct_buff.species.sort(key=itemgetter(1), reverse=True)
        struct_buff.species.sort(key=itemgetter(0))

    
        #Replace sites
        for i in range(n):
            struct_buff.sites[final_species[i]]=struct_buff.sites.pop(init_species[i])
        
        #And sort
        species_strs = struct_buff._format_style("{ele}{charge}{sign}").split(" ")
        struct_buff.sites={spec: struct_buff.sites[spec] for spec in species_strs}
    
        return struct_buff

    def sub_prob(self, s1: str, s2: str) -> float:
        """Calculate the probability of substitution of two species."""
        return np.exp(self.get_lambda(s1, s2)) / self.Z

    def sub_probs(self, s1: str) -> pd.Series:
        """Determine the substitution probabilities of a species with others.

        Determines the probability of substitution of the species with every
        species in the lambda table.

        """
        probs = self.get_lambdas(s1)
        probs = np.exp(probs)
        probs /= self.Z
        return probs

    def complete_sub_probs(self) -> pd.DataFrame:
        """Generate a DataFrame with all the substitution probabilities."""
        return np.exp(self.lambda_tab) / self.Z

    def complete_cond_probs(self) -> pd.DataFrame:
        """Generate a DataFrame with all the conditional substitution probabilities."""
        lambda_exp = np.exp(self.lambda_tab)
        return lambda_exp / lambda_exp.sum(axis=1)

    def complete_pair_corrs(self) -> pd.DataFrame:
        """Generate a DataFrame with all the pair correlations."""
        corr = self.complete_sub_probs()

        # Sum each row (symmetry means this is the same as column sums)
        sums = corr.sum(axis=0)
        # Stack into matrix
        mat_sums = np.vstack([sums] * len(sums))
        # Make each element the product of (row_sum * col_sum)
        mat_sums *= sums[:, None]

        corr /= mat_sums

        return corr

    def same_spec_probs(self) -> pd.Series:
        """Calculate the same species substiution probabilities."""
        return np.exp(
          pd.Series(
            np.diag(self.lambda_tab),
            index=[self.lambda_tab.index, self.lambda_tab.columns], )
        ) / self.Z

    def same_spec_cond_probs(self) -> pd.Series:
        """Calculate the same species conditional substiution probabilities."""
        return (
          np.exp(self.lambda_tab.to_numpy().diagonal()) / np.exp(self.lambda_tab).sum(axis=0)
        )

    def pair_corr(self, s1: str, s2: str) -> float:
        """Determine the pair correlation of two ionic species."""
        corr = self.sub_prob(s1, s2)
        corr /= self.sub_probs(s1).sum()
        corr /= self.sub_probs(s2).sum()
        return corr

    def cond_sub_prob(self, s1: str, s2: str) -> float:
        """Calculate the probability of substitution of one species with another."""
        return np.exp(self.get_lambda(s1, s2)) / np.exp(self.get_lambdas(s2)).sum()

    def cond_sub_probs(self, s1: str) -> pd.Series:
        """Calculate the probabilities of substitution of a given species.

        Calculates probabilities of substitution of given species with all
        others in the lambda table.

        """
        probs = self.get_lambdas(s1)
        probs = np.exp(probs)
        probs /= np.exp(self.lambda_tab).sum()
        return probs

    def unary_substitute(
      self,
      structure: SmactStructure,
      thresh: Optional[float] = 1e-5, ) -> Generator[Tuple[SmactStructure, float], None, None]:
        """Find all structures with 1 substitution with probability above a threshold.

        Args:
            structure: A :class:`SmactStructure` instance from which to generate compounds.
            thresh (float): The probability threshold; discard all substitutions that have
                a probability to generate a naturally-occuring compound less than this.

        Yields:
            Tuples of (:class:`SmactStructure`, probability).

        """
        for specie in structure.get_spec_strs():
            cond_probs = self.cond_sub_probs(specie)
            likely_probs = cond_probs.loc[cond_probs > thresh]

            for new_spec, prob in likely_probs.items():
                if any([
                  new_spec == specie,
                  parse_spec(specie)[1] != parse_spec(new_spec)[1], ]):
                    continue
                yield (self._mutate_structure(structure, specie, new_spec), prob)



"""Probability models for species substitution.

Implements base class :class:`SubstitutionModel`,
which can be extended to allow for development of new
lambda tables. An example of such an extension,
:class:`RadiusModel`, is also implemented.

Todo:
    * Allow for parallelism in lambda table calculations
      by implementing a `sub_probs` abstractmethod
      that :meth:`SubstitutionModel.gen_lambda` uses,
      if available.

"""

import abc
import os
from itertools import combinations_with_replacement
from typing import Dict, List, Optional

import pandas as pd

from smact import data_directory

from .utilities import parse_spec


class SubstitutionModel(abc.ABC):
    """Abstract base class for substitution models."""

    @abc.abstractmethod
    def sub_prob(self, s1: str, s2: str) -> float:
        """Calculate the probability of substituting species s1 for s2.

        Args:
            s1: The species being substituted.
            s2: The species substituting.

        Returns:
            The probability of substitution.

        """

    def gen_lambda(self, species: List[str]) -> pd.DataFrame:
        """Generate a lambda table for a list of species.

        Args:
            species: A list of species strings.

        Returns:
            A pivot table-style DataFrame containing lambda values
            for every possible species pair.

        """
        pairs = combinations_with_replacement(species, 2)

        lambda_tab = []
        for s1, s2 in pairs:
            prob = self.sub_prob(s1, s2)
            lambda_tab.append((s1, s2, prob))
            if s1 != s2:
                lambda_tab.append((s2, s1, prob))

        df = pd.DataFrame(lambda_tab)
        return df.pivot(index=0, columns=1, values=2)


class RadiusModel(SubstitutionModel):
    """Substitution probability model based on Shannon radii."""

    def __init__(self):
        r"""Parse Shannon radii data file.

        Also calculates "spring constant", _k_, based on maximum
        difference in Shannon radii:

        .. math::
            k = \Delta r_\mathrm{max}^{-2}.

        """
        shannon_file = os.path.join(data_directory, "shannon_radii.csv")

        self.shannon_data = pd.read_csv(shannon_file, index_col=0)

        self.k = (
          self.shannon_data["ionic_radius"].max() - self.shannon_data["ionic_radius"].min()
        )**-2

    def sub_prob(self, s1, s2):
        r"""Calculate the probability of substituting species s1 for s2.

        Based on the difference in Shannon radii, the probability is
        assumed to be:

        .. math::
            p = 1 - k \Delta r^2.

        Args:
            s1: The species being substituted.
            s2: The species substituting.

        Returns:
            The probability of substitution.

        """
        spec1 = parse_spec(s1)
        spec2 = parse_spec(s2)

        try:
            ele1_rows = self.shannon_data.loc[spec1[0]]
            ele2_rows = self.shannon_data.loc[spec2[0]]
        except KeyError as e:
            raise KeyError(f"Element not in Shannon radius data file: {e}")

        spec1_rows = ele1_rows.loc[ele1_rows["charge"] == spec1[1]]
        spec2_rows = ele2_rows.loc[ele2_rows["charge"] == spec2[1]]

        # Get mean so we don't need coordination information
        mean_spec1_r = spec1_rows["ionic_radius"].mean()
        mean_spec2_r = spec2_rows["ionic_radius"].mean()

        # Hooke's law-style probability
        return 1 - self.k * (mean_spec1_r - mean_spec2_r)**2



from builtins import map
from builtins import input
from builtins import range
import smact
from numpy import sqrt, product

def eneg_mulliken(element):
    """Get Mulliken electronegativity from the IE and EA.

    Arguments:
        symbol (smact.Element or str): Element object or symbol

    Returns:
        mulliken (float): Mulliken electronegativity

    """
    if type(element) == str:
        element = smact.Element(element)
    elif type(element) != smact.Element:
        raise Exception("Unexpected type: {0}".format(type(element)))

    mulliken = (element.ionpot+element.e_affinity)/2.0

    return mulliken


def band_gap_Harrison(anion, cation, verbose=False,
                      distance=None):

    """
    Estimates the band gap from elemental data.

    The band gap is estimated using the principles outlined in
    Harrison's 1980 work "Electronic Structure and the Properties of
    Solids: The Physics of the Chemical Bond".

    Args:
        Anion (str): Element symbol of the dominant anion in the system

        Cation (str): Element symbol of the the dominant cation in the system
        Distance (float or str): Nuclear separation between anion and cation
                i.e. sum of ionic radii
        verbose (bool) : An optional True/False flag. If True, additional
        information is printed to the standard output. [Defult: False]

    Returns :
        Band_gap (float): Band gap in eV

    """

    # Set constants
    hbarsq_over_m = 7.62

    # Get anion and cation
    An = anion
    Cat = cation
    d = float(distance)

    # Get elemental data:
    elements_dict = smact.element_dictionary((An, Cat))
    An, Cat = elements_dict[An], elements_dict[Cat]

    # Calculate values of equation components
    V1_Cat = (Cat.eig - Cat.eig_s)/4
    V1_An = (An.eig - An.eig_s)/4
    V1_bar = (V1_An + V1_Cat)/2
    V2 = 2.16 * hbarsq_over_m / (d**2)
    V3 = (Cat.eig - An.eig)/2
    alpha_m = (1.11*V1_bar)/sqrt(V2**2 + V3**2)

    # Calculate Band gap [(3-43) Harrison 1980 ]
    Band_gap = (3.60/3.)*(sqrt(V2**2 + V3**2))*(1-alpha_m)
    if verbose:
        print("V1_bar = ", V1_bar)
        print("V2 = ", V2)
        print("alpha_m = ", alpha_m)
        print("V3 = ", V3)

    return Band_gap


def compound_electroneg(verbose=False, elements=None, stoichs=None,
                                                source='Mulliken'):

    """Estimate electronegativity of compound from elemental data.

    Uses Mulliken electronegativity by default, which uses elemental
    ionisation potentials and electron affinities. Alternatively, can
    use Pauling electronegativity, re-scaled by factor 2.86 to achieve
    same scale as Mulliken method (Nethercot, 1974)
    DOI:10.1103/PhysRevLett.33.1088 .

    Geometric mean is used (n-th root of product of components), e.g.:

    X_Cu2S = (X_Cu * X_Cu * C_S)^(1/3)

    Args:
        elements (list) : Elements given as standard elemental symbols.
        stoichs (list) : Stoichiometries, given as integers or floats.
        verbose (bool) : An optional True/False flag. If True, additional information
            is printed to the standard output. [Default: False]
        source: String 'Mulliken' or 'Pauling'; type of Electronegativity to
            use. Note that in SMACT, Pauling electronegativities are
            rescaled to a Mulliken-like scale.

    Returns:
        Electronegativity (float) : Estimated electronegativity (no units).

    """
    if type(elements[0]) == str:
        elementlist = [smact.Element(i) for i in elements]
    elif type(elements[0]) == smact.Element:
        elementlist = elements
    else:
        raise Exception("Please supply a list of element symbols or SMACT Element objects")

    stoichslist = stoichs
    # Convert stoichslist from string to float
    stoichslist = list(map(float, stoichslist))

    # Get electronegativity values for each element

    if source == 'Mulliken':
        elementlist = [(el.ionpot+el.e_affinity)/2.0
                       for el in elementlist]

    elif source == 'Pauling':
        elementlist = [(2.86 * el.pauling_eneg)
                       for el in elementlist]
    else:
        raise Exception("Electronegativity type '{0}'".format(source),
                        "is not recognised")

    # Print optional list of element electronegativities.
    # This may be a useful sanity check in case of a suspicious result.
    if verbose:
        print("Electronegativities of elements=", elementlist)

    # Raise each electronegativity to its appropriate power
    # to account for stoichiometry.
    for i in range(0, len(elementlist)):
        elementlist[i] = [elementlist[i]**stoichslist[i]]

    # Calculate geometric mean (n-th root of product)
    prod = product(elementlist)
    compelectroneg = (prod)**(1.0/(sum(stoichslist)))

    if verbose:
        print("Geometric mean = Compound 'electronegativity'=", compelectroneg)

    return compelectroneg



"""Minimalist structure representation for comprehensible manipulation."""

import logging
import re
from collections import defaultdict
from operator import itemgetter
from typing import Dict, List, Optional, Tuple, Union

from math import gcd
from functools import reduce

import numpy as np
import pymatgen
from pymatgen.analysis.bond_valence import BVAnalyzer
from pymatgen.ext.matproj import MPRester

import smact

from . import logger
from .utilities import get_sign

class SmactStructure:
    """SMACT implementation inspired by pymatgen Structure class.

    Handles basic structural and compositional information for a compound.
    Includes a lossless POSCAR-style specification for storing structures,
    allowing structures to be stored in files or databases, or to be pulled
    from the `Materials Project <https://www.materialsproject.org>`_.

    Attributes:
        species: A list of tuples describing the composition of the structure,
            stored as (element, oxidation, stoichiometry). The list is sorted
            alphabetically based on element symbol, and identical elements
            are sorted with highest charge first.
        lattice_mat: A numpy 3x3 array containing the lattice vectors.
        sites: A dictionary of {species: coords}, where species is a string
            representation of the species and coords is a list of position
            vectors, given as lists of length 3. For example:

            >>> s = SmactStructure.from_file('tests/files/NaCl.txt')
            >>> s.sites
            {'Cl1-': [[2.323624165, 1.643050405, 4.02463512]], 'Na1+': [[0.0, 0.0, 0.0]]}

        lattice_param: The lattice parameter.

    """

    def __init__(
      self,
      species: List[Union[Tuple[str, int, int], Tuple[smact.Species, int]]],
      lattice_mat: np.ndarray,
      sites: Dict[str, List[List[float]]],
      lattice_param: Optional[float] = 1.0,
      sanitise_species: Optional[bool] = True,
    ):
        """Initialize structure with constituent species.

        Args:
            species: See :class:`~.SmactStructure`. May be supplied as either a list of
                (element, oxidation, stoichiometry) or (:class:`~smact.Species`, stoichiometry).
            lattice_mat: See :class:`~.SmactStructure`.
            sites: See :class:`~.SmactStructure`.
            lattice_param: See :class:`~.SmactStructure`.
            sanitise_species: Whether to sanitise check species. Should be `True` unless
                species have already been sanitised by a different constructor like
                :meth:`~.from_mp`.

        """
        self.species = self._sanitise_species(species) if sanitise_species else species

        self.lattice_mat = lattice_mat

        self.sites = {spec: sites[spec] for spec in self.get_spec_strs()}  # Sort sites

        self.lattice_param = lattice_param

    def __repr__(self):
        """Represent the structure as a POSCAR.

        Alias for :meth:`~.as_poscar`.

        """
        return self.as_poscar()

    def __eq__(self, other):
        """Determine equality of SmactStructures based on their attributes.

        :attr:`~.species`, :attr:`~.lattice_mat`, :attr:`~.lattice_param` and
        :attr:`~.sites` must all be equal for the comparison to be True.

        Note:
            For the SmactStructures to be equal their attributes must be
            *identical*. For example, it is insufficient that the two
            structures have the same space group or the same species;
            the site coordinates must be equal also.

        """
        if not isinstance(other, SmactStructure):
            return False
        return all([
          self.species == other.species,
          np.array_equal(self.lattice_mat, other.lattice_mat),
          self.lattice_param == other.lattice_param,
          self.sites == other.sites,
          list(self.sites.keys()) == list(other.sites.keys()),
        ])

    @staticmethod
    def _sanitise_species(
      species: List[Union[Tuple[str, int, int], Tuple[smact.Species, int]]],
    ) -> List[Tuple[str, int, int]]:
        """Sanitise and format a list of species.

        Args:
            species: See :meth:`~.__init__`.

        Returns:
            sanit_species: Sanity-checked species in the format of
            a list of (element, oxidation, stoichiometry).

        Raises:
            TypeError: species contains the wrong types.
            ValueError: species is either empty or contains tuples of
                incorrect length.

        """
        if not isinstance(species, list):
            raise TypeError(f"`species` must be a list, got {type(species)}.")
        if len(species) == 0:
            raise ValueError("`species` cannot be empty.")
        if not isinstance(species[0], tuple):
            raise TypeError(f"`species` must be a list of tuples, got list of {type(species[0])}.")

        species_error = (
          "`species` list of tuples must contain either "
          "2-tuples of Species objects and stoichiometries, "
          "or 3-tuples of elements, oxidations and stoichiometries."
        )
        if len(species[0]) not in {2, 3}:
            raise ValueError(species_error)

        if isinstance(species[0][0], str):  # String variation of instantiation
            species.sort(key=itemgetter(1), reverse=True)
            species.sort(key=itemgetter(0))
            sanit_species = species

        elif isinstance(species[0][0], smact.Species):  # Species class variation of instantiation
            species.sort(key=lambda x: (x[0].symbol, -x[0].oxidation))
            sanit_species = [(x[0].symbol, x[0].oxidation, x[1]) for x in species]

        else:
            raise TypeError(species_error)

        return sanit_species

    @staticmethod
    def __parse_py_sites(
      structure: pymatgen.core.Structure,
    ) -> Tuple[Dict[str, List[List[float]]], List[Tuple[str, int, int]]]:
        """Parse the sites of a pymatgen Structure.

        Args:
            structure: A :class:`pymatgen.core.Structure` instance.

        Returns:
            sites (dict): In which a key is a species string
                and its corresponding value is a list of the coordinates
                that species occupies in the supercell. The coordinates
                are represented by lists containing three elements: one
                for each spatial dimension.
            species (list): A list of each species in the structure,
                represented by a tuple of (element, charge, stoichiometry).

        """
        if not isinstance(structure, pymatgen.core.Structure):
            raise TypeError("structure must be a pymatgen.core.Structure instance.")

        sites = defaultdict(list)
        for site in structure.sites:
            site_type = site.species_string
            # Add charge magnitude, for cases of unit charge
            if all([
              site_type[-2] not in map(str, range(10)),
              site_type[-1] in ("+", "-"), ]):
                site_type = site_type[:-1] + '1' + site_type[-1]

            sites[site_type].append(site.coords.tolist())

        sites = dict(sites)

        # Find stoichiometry
        total_specs = [len(val) for val in sites.values()]
        hcf = reduce(gcd, total_specs)
        total_specs = [int(x / hcf) for x in total_specs]

        species = []
        for spec, stoic in zip(sites.keys(), total_specs):
            charge_match = re.search(r"\d+", spec)

            if charge_match:
                charge_loc = charge_match.start()
                symb = spec[:charge_loc]
                charge = int(spec[-1] + spec[charge_loc:-1])
            else:
                symb = spec
                charge = 0

            species.append((symb, charge, stoic))

        return sites, species

    @staticmethod
    def from_py_struct(structure: pymatgen.core.Structure):
        """Create a SmactStructure from a pymatgen Structure object.

        Args:
            structure: A pymatgen Structure.

        Returns:
            :class:`~.SmactStructure`

        """
        if not isinstance(structure, pymatgen.core.Structure):
            raise TypeError("Structure must be a pymatgen.core.Structure instance.")

        bva = BVAnalyzer()
        struct = bva.get_oxi_state_decorated_structure(structure)

        sites, species = SmactStructure.__parse_py_sites(struct)

        lattice_mat = struct.lattice.matrix

        lattice_param = 1.0

        return SmactStructure(
          species,
          lattice_mat,
          sites,
          lattice_param,
          sanitise_species=True, )

    @staticmethod
    def from_mp(
      species: List[Union[Tuple[str, int, int], Tuple[smact.Species, int]]],
      api_key: str, ):
        """Create a SmactStructure using the first Materials Project entry for a composition.

        Args:
            species: See :meth:`~.__init__`.
            api_key: A www.materialsproject.org API key.

        Returns:
            :class:`~.SmactStructure`

        """
        sanit_species = SmactStructure._sanitise_species(species)

        with MPRester(api_key) as m:
            eles = SmactStructure._get_ele_stoics(sanit_species)
            formula = "".join(f"{ele}{stoic}" for ele, stoic in eles.items())
            structs = m.query(
              criteria={"reduced_cell_formula": formula},
              properties=["structure"], )

            if len(structs) == 0:
                raise ValueError(
                  "Could not find composition in Materials Project Database, "
                  "please supply a structure."
                )

            struct = structs[0]['structure']  # Default to first found structure

        if 0 not in (spec[1] for spec in sanit_species):  # If everything's charged
            bva = BVAnalyzer()
            struct = bva.get_oxi_state_decorated_structure(struct)

        lattice_mat = struct.lattice.matrix

        lattice_param = 1.0  # TODO Use actual lattice parameter

        sites, _ = SmactStructure.__parse_py_sites(struct)

        return SmactStructure(
          sanit_species,
          lattice_mat,
          sites,
          lattice_param,
          sanitise_species=False, )

    @staticmethod
    def from_file(fname: str):
        """Create SmactStructure from a POSCAR file.

        Args:
            fname: The name of the POSCAR file.
                See :meth:`~.as_poscar` for format specification.

        Returns:
            :class:`~.SmactStructure`

        """
        with open(fname, 'r') as f:
            return SmactStructure.from_poscar(f.read())

    @staticmethod
    def from_poscar(poscar: str):
        """Create SmactStructure from a POSCAR string.

        Args:
            poscar: A SMACT-formatted POSCAR string.
                See :meth:`~.as_poscar` for format specification.

        Returns:
            :class:`~.SmactStructure`

        """
        lines = poscar.split("\n")

        # Find stoichiometry
        total_specs = [int(x) for x in lines[6].split(" ")]
        hcf = reduce(gcd, total_specs)
        total_specs = [int(x / hcf) for x in total_specs]

        species = []
        for spec_str, stoic in zip(lines[0].split(" "), total_specs):
            charge_match = re.search(r"\d+", spec_str)

            if charge_match:
                charge_loc = charge_match.start()
                symb = spec_str[:charge_loc]
                charge = int(spec_str[-1] + spec_str[charge_loc:-1])
            else:
                symb = spec_str
                charge = 0

            species.append((symb, charge, stoic))

        lattice_param = float(lines[1])

        lattice = np.array([[float(point) for point in line.split(" ")] for line in lines[2:5]])

        sites = defaultdict(list)
        for line in lines[8:]:
            if not line:  # EOF
                break

            split_line = line.split(" ")
            coords = [float(x) for x in split_line[:3]]
            spec = split_line[-1]

            sites[spec].append(coords)

        sites = dict(sites)

        return SmactStructure(species, lattice, sites, lattice_param)

    def _format_style(
      self,
      template: str,
      delim: Optional[str] = " ",
      include_ground: Optional[bool] = False, ) -> str:
        """Format a given template string with the composition.

        Formats a python template string with species information,
        with each species separated by a given delimiter.

        Args:
            template: Template string to format, using python's
                curly brackets notation. Supported keywords are
                `ele` for the elemental symbol, `stoic` for the
                stoichiometry, `charge` for the absolute value
                of oxidation state and `sign` for the 
                oxidation state's sign.
            delim: The delimeter between species' templates.
            include_ground: Whether to include the charge and sign
                of neutral species.
        
        Returns:
            String of templates formatted for each species, separated
                by `delim`.
        
        Examples:
            >>> s = SmactStructure.from_file('tests/files/CaTiO3.txt')
            >>> template = '{stoic}x{ele}{charge}{sign}'
            >>> print(s._format_style(template))
            1xCa2+ 3xO2- 1xTi4+

        """
        if include_ground:
            return delim.join(
              template.format(
                ele=specie[0],
                stoic=specie[2],
                charge=abs(specie[1]),
                sign="+" if specie[1] >= 0 else "-",
              ) for specie in self.species
            )

        return delim.join(
          template.format(
            ele=specie[0],
            stoic=specie[2],
            charge=abs(specie[1]) if specie[1] != 0 else "",
            sign=get_sign(specie[1]),
          ) for specie in self.species
        )

    @staticmethod
    def _get_ele_stoics(species: List[Tuple[str, int, int]]) -> Dict[str, int]:
        """Get the number of each element type in the compound, irrespective of oxidation state.

        Args:
            species: See :meth:`~.__init__`.

        Returns:
            eles: Dictionary of {element: stoichiometry}.

        Examples:
            >>> species = [('Fe', 2, 1), ('Fe', 3, 2), ('O', -2, 4)]
            >>> print(SmactStructure._get_ele_stoics(species))
            {'Fe': 3, 'O': 4}

        """
        eles = defaultdict(int)
        for specie in species:
            eles[specie[0]] += specie[2]

        return dict(eles)

    def has_species(self, species: Tuple[str, int]) -> bool:
        """Determine whether a given species is in the structure."""
        return species in map(itemgetter(0, 1), self.species)

    def get_spec_strs(self) -> List[str]:
        """Get string representations of the constituent species.

        Returns:
            A list of strings, formatted as '{element}{charge}{sign}'.
        
        Examples:
            >>> s = SmactStructure.from_file('tests/files/CaTiO3.txt')
            >>> s.get_spec_strs()
            ['Ca2+', 'O2-', 'Ti4+']
        
        """
        return self._format_style("{ele}{charge}{sign}").split(' ')

    def composition(self) -> str:
        """Generate a key that describes the composition.

        Key format is '{element}_{stoichiometry}_{charge}{sign}'
        with no delimiter, *sans brackets*. Species are ordered as stored within
        the structure, see :class:`~.SmactStructure`.

        Returns:
            Key describing constituent species.

        Examples:
            >>> s = SmactStructure.from_file('tests/files/CaTiO3.txt')
            >>> print(s.composition())
            Ca_1_2+O_3_2-Ti_1_4+

        """
        comp_style = "{ele}_{stoic}_{charge}{sign}"
        return self._format_style(comp_style, delim="", include_ground=True)

    def as_poscar(self) -> str:
        """Represent the structure as a POSCAR file compatible with VASP5.

        The POSCAR format adopted is as follows:

        The first line contains the species' names separated by a whitespace.
        The second through fourth line, inclusive, contain the lattice
        matrix: each line contains a lattice vector, with elements
        separated by a whitespace.
        The fifth line contains the elements' names separated by a whitespace.
        If more than one oxidation state exists for an element, the element
        appears multiple times; once for each oxidation state.
        The sixth line is the string 'Cartesian'.
        The seventh line onwards are the Cartesian coordinates of each site,
        separated by a whitespace. In addition, at the end of each line is the
        species' name, separated by a whitespace.

        For examples of this format, see the text files under tests/files.

        Returns:
            str: POSCAR-style representation of the structure.

        """
        poscar = " ".join(self.get_spec_strs()) + "\n"

        poscar += f"{self.lattice_param}\n"

        poscar += "\n".join(" ".join(map(str, vec)) for vec in self.lattice_mat.tolist()) + "\n"

        spec_count = {spec: len(coords) for spec, coords in self.sites.items()}

        poscar += self._format_style("{ele}") + "\n"

        poscar += " ".join(str(spec_count[spec]) for spec in self.get_spec_strs()) + "\n"

        poscar += "Cartesian\n"
        for spec, coords in self.sites.items():
            for coord in coords:
                poscar += " ".join(map(str, coord))
                poscar += f" {spec}\n"

        return poscar



"""
Provide data from text files while transparently caching for efficiency.

This module handles the loading of external data used to initialise the
core smact.Element and smact.Species classes.  It implements a
transparent data-caching system to avoid a large amount of I/O when
naively constructing several of these objects.  It also implements a
switchable system to print verbose warning messages about possible
missing data (mainly for debugging purposes). In general these fuctions
are used in the background and it is not necessary to use them directly.
"""

from builtins import next
from builtins import map
from builtins import zip
import csv
import os

from smact import data_directory

# Module-level switch: print "verbose" warning messages
# about missing data.
_print_warnings = False


def set_warnings(enable=True):
    """Set verbose warning messages on and off.

    In order to see any of the warnings, this function needs to be
    called _before_ the first call to the smact.Element()
    constructor.

    Args:
    enable (bool) : print verbose warning messages.
    """

    global _print_warnings
    _print_warnings = enable


def _get_data_rows(filename):
    """Generator for datafile entries by row"""
    with open(filename, 'r') as file:
        for line in file:
            line = line.strip()
            if line[0] != '#':
                yield line.split()

def float_or_None(x):
    """Cast a string to a float or to a None."""
    try:
        return float(x)
    except ValueError:
        return None

# Loader and cache for the element oxidation-state data.
_el_ox_states = None


def lookup_element_oxidation_states(symbol, copy=True):
    """
    Retrieve a list of known oxidation states for an element.
    The oxidation states list used is the SMACT default and
    most exhaustive list.

    Args:
        symbol (str) : the atomic symbol of the element to look up.
        copy (Optional(bool)): if True (default), return a copy of the
            oxidation-state list, rather than a reference to the cached
            data -- only use copy=False in performance-sensitive code
            and where the list will not be modified!

    Returns:
        list: List of known oxidation states for the element.

            Returns None if oxidation states for the Element were not
            found in the external data.
    """

    global _el_ox_states

    if _el_ox_states is None:
        _el_ox_states = {}

        for items in _get_data_rows(os.path.join(data_directory,
                                                 "oxidation_states.txt")):
            _el_ox_states[items[0]] = [int(oxidationState)
                                       for oxidationState in items[1:]]

    if symbol in _el_ox_states:
        if copy:
            # _el_ox_states stores lists -> if copy is set, make an implicit
            # deep copy.  The elements of the lists are integers, which are
            # "value types" in Python.

            return [oxidationState for oxidationState in _el_ox_states[symbol]]
        else:
            return _el_ox_states[symbol]
    else:
        if _print_warnings:
            print("WARNING: Oxidation states for element {0} "
                  "not found.".format(symbol))
        return None

_el_ox_states_icsd = None


def lookup_element_oxidation_states_icsd(symbol, copy=True):
    """
    Retrieve a list of known oxidation states for an element.
    The oxidation states list used contains only those found
    in the ICSD (and judged to be non-spurious).

    Args:
        symbol (str) : the atomic symbol of the element to look up.
        copy (Optional(bool)): if True (default), return a copy of the
            oxidation-state list, rather than a reference to the cached
            data -- only use copy=False in performance-sensitive code
            and where the list will not be modified!

    Returns:
        list: List of known oxidation states for the element.

            Return None if oxidation states for the Element were not
            found in the external data.
    """

    global _el_ox_states_icsd

    if _el_ox_states_icsd is None:
        _el_ox_states_icsd = {}

        for items in _get_data_rows(os.path.join(data_directory,
                                                 "oxidation_states_icsd.txt")):
            _el_ox_states_icsd[items[0]] = [int(oxidationState)
                                       for oxidationState in items[1:]]
    if symbol in _el_ox_states_icsd:
        if copy:
            # _el_ox_states_icsd stores lists -> if copy is set, make an implicit
            # deep copy. The elements of the lists are integers, which are
            # "value types" in Python.
            return [oxidationState for oxidationState in _el_ox_states_icsd[symbol]]
        else:
            return _el_ox_states_icsd[symbol]
    else:
        if _print_warnings:
            print("WARNING: Oxidation states for element {0}"
            "not found.".format(symbol))
        return None

_el_ox_states_sp = None

def lookup_element_oxidation_states_sp(symbol, copy=True):
    """
    Retrieve a list of known oxidation states for an element.
    The oxidation states list used contains only those that
    are in the Pymatgen default lambda table for structure prediction.

    Args:
        symbol (str) : the atomic symbol of the element to look up.
        copy (Optional(bool)): if True (default), return a copy of the
            oxidation-state list, rather than a reference to the cached
            data -- only use copy=False in performance-sensitive code
            and where the list will not be modified!

    Returns:
        list: List of known oxidation states for the element.

            Return None if oxidation states for the Element were not
            found in the external data.
    """

    global _el_ox_states_sp

    if _el_ox_states_sp is None:
        _el_ox_states_sp = {}

        for items in _get_data_rows(os.path.join(data_directory,
                                                 "oxidation_states_SP.txt")):
            _el_ox_states_sp[items[0]] = [int(oxidationState)
                                       for oxidationState in items[1:]]


    if symbol in _el_ox_states_sp:
        if copy:
            # _el_ox_states_sp stores lists -> if copy is set, make an implicit
            # deep copy.  The elements of the lists are integers, which are
            # "value types" in Python.

            return [oxidationState for oxidationState in _el_ox_states_sp[symbol]]
        else:
            return _el_ox_states_sp[symbol]
    else:
        if _print_warnings:
            print("WARNING: Oxidation states for element {0} "
                  "not found.".format(symbol))
        return None

# Loader and cache for the element HHI scores.

_element_hhis = None


def lookup_element_hhis(symbol):
    """
    Retrieve the HHI_R and HHI_p scores for an element.

    Args:
        symbol : the atomic symbol of the element to look up.

    Returns:
        tuple : (HHI_p, HHI_R)

            Return None if values for the elements were
            not found in the external data.
    """

    global _element_hhis

    if _element_hhis is None:
        _element_hhis = {}

        with open(os.path.join(data_directory, "HHIs.txt"),
                  'r') as file:
            for line in file:
                line = line.strip()

                if line[0] != '#':
                    items = line.split()

                    _element_hhis[items[0]] = (float(items[1]),
                                               float(items[2]))

    if symbol in _element_hhis:
        return _element_hhis[symbol]
    else:
        if _print_warnings:
            print("WARNING: HHI data for element "
                  "{0} not found.".format(symbol))

        return None

# Loader and cache for elemental data

_element_data = None

def lookup_element_data(symbol, copy=True):
    """
    Retrieve tabulated data for an element.

    The table "data/element_data.txt" contains a collection of relevant
    atomic data. If a cache exists in the form of the module-level
    variable _element_data, this is returned. Otherwise, a dictionary is
    constructed from the data table and cached before returning it.

    Args:
        symbol (str) : Atomic symbol for lookup

        copy (Optional(bool)) : if True (default), return a copy of the
            data dictionary, rather than a reference to the cached
            object -- only used copy=False in performance-sensitive code
            and where you are certain the dictionary will not be
            modified!

    Returns (dict) : Dictionary of data for given element, keyed by
        column headings from data/element_data.txt.
    """
    global _element_data
    if _element_data is None:
        _element_data = {}
        keys = ('Symbol', 'Name', 'Z', 'Mass', 'r_cov', 'e_affinity',
                'p_eig', 's_eig', 'Abundance', 'el_neg', 'ion_pot',
                'dipol')
        for items in _get_data_rows(os.path.join(data_directory,
                                                 "element_data.txt")):
            # First two columns are strings and should be left intact
            # Everything else is numerical and should be cast to a float
            # or, if not clearly a number, to None
            clean_items = items[0:2] + list(map(float_or_None, items[2:]))

            _element_data.update({items[0]:
                                  dict(list(zip(keys, clean_items)))})

    if symbol in _element_data:
        if copy:
            # _element_open_babel_derived_data stores dictionaries
            # -> if copy is set, use the dict.copy() function to return
            # a copy. The values are all Python "value types", so
            # explicitly cloning the elements is not necessary to make
            # a deep copy.

            return _element_data[symbol].copy()
        else:
            return _element_data[symbol]
    else:
        if _print_warnings:
            print("WARNING: Elemental data for {0}"
                  " not found.".format(symbol))
            print(_element_data)
        return None


# Loader and cache for the element Shannon radii datasets.

_element_shannon_radii_data = None


def lookup_element_shannon_radius_data(symbol, copy=True):
    """
    Retrieve Shannon radii for known states of an element.

    Retrieve Shannon radii for known oxidation states and coordination
    environments of an element.

    Args:
        symbol (str) : the atomic symbol of the element to look up.

        copy (Optional(bool)): if True (default), return a copy of the data
        dictionary, rather than a reference to the cached object --
        only use copy=False in performance-sensitive code and where
        you are certain the dictionary will not be modified!

    Returns:
        list:
            Shannon radii datasets.

        Returns None if the element was not found among the external
        data.

        Shannon radii datasets are dictionaries with the keys:

        charge
            *int* charge
        coordination
            *int* coordination
        crystal_radius
            *float*
        ionic_radius
            *float*
        comment
            *str*
    """

    global _element_shannon_radii_data

    if _element_shannon_radii_data is None:
        _element_shannon_radii_data = {}

        with open(os.path.join(data_directory, "shannon_radii.csv"),
                  'r') as file:
            reader = csv.reader(file)

            # Skip the first row (headers).

            next(reader)

            for row in reader:
                # For the shannon radii, there are multiple datasets for
                # different element/oxidation-state/coordination
                # combinations.

                key = row[0]

                dataset = {
                    'charge': int(row[1]),
                    'coordination': row[2],
                    'crystal_radius': float(row[3]),
                    'ionic_radius': float(row[4]),
                    'comment': row[5]
                    }

                if key in _element_shannon_radii_data:
                    _element_shannon_radii_data[key].append(dataset)
                else:
                    _element_shannon_radii_data[key] = [dataset]

    if symbol in _element_shannon_radii_data:
        if copy:
            # _element_shannon_radii_data stores a list of dictionaries
            # -> if copy is set, copy the list and use the dict.copy()
            # function on each element.
            # The dictionary values are all Python "value types", so
            # nothing further is required to make a deep copy.
            return [item.copy() for item in
                    _element_shannon_radii_data[symbol]]
        else:
            return _element_shannon_radii_data[symbol]
    else:
        if _print_warnings:
            print("WARNING: Shannon-radius data for element {0} not "
                  "found.".format(symbol))

        return None

# Loader and cache for the machine-learned extended element Shannon radii datasets.

_element_shannon_radii_data_extendedML = None


def lookup_element_shannon_radius_data_extendedML(symbol, copy=True):
    """
    Retrieve the machine learned extended Shannon radii for
    known states of an element.


    Retrieve Shannon radii for known oxidation states and coordination
    environments of an element.

    Source of extended radii is:
    Baloch, A.A., Alqahtani, S.M., Mumtaz, F., Muqaibel, A.H., Rashkeev,
    S.N. and Alharbi, F.H., 2021.
    Extending Shannon's Ionic Radii Database Using Machine Learning.
    arXiv preprint arXiv:2101.00269.

    Args:
        symbol (str) : the atomic symbol of the element to look up.

        copy (Optional(bool)): if True (default), return a copy of the data
        dictionary, rather than a reference to the cached object --
        only use copy=False in performance-sensitive code and where
        you are certain the dictionary will not be modified!

    Returns:
        list:
            Extended Shannon radii datasets.

        Returns None if the element was not found among the external
        data.

        Shannon radii datasets are dictionaries with the keys:

        charge
            *int* charge
        coordination
            *int* coordination
        ionic_radius
            *float*
        comment
            *str*
    """

    global _element_shannon_radii_data_extendedML

    if _element_shannon_radii_data_extendedML is None:
        _element_shannon_radii_data_extendedML = {}

        with open(os.path.join(data_directory, "shannon_radii_ML_extended.csv"),
                  'r') as file:
            reader = csv.reader(file)

            # Skip the first row (headers).

            next(reader)

            for row in reader:
                # For the shannon radii, there are multiple datasets for
                # different element/oxidation-state/coordination
                # combinations.

                key = row[0]

                dataset = {
                    'charge': int(row[1]),
                    'coordination': row[2],
                    'crystal_radius': float(row[3]),
                    'ionic_radius': float(row[4]),
                    'comment': row[5]
                    }

                if key in _element_shannon_radii_data_extendedML:
                    _element_shannon_radii_data_extendedML[key].append(dataset)
                else:
                    _element_shannon_radii_data_extendedML[key] = [dataset]

    if symbol in _element_shannon_radii_data_extendedML:
        if copy:
            # _element_shannon_radii_data_extendedML stores a list of dictionaries
            # -> if copy is set, copy the list and use the dict.copy()
            # function on each element.
            # The dictionary values are all Python "value types", so
            # nothing further is required to make a deep copy.
            return [item.copy() for item in
                    _element_shannon_radii_data_extendedML[symbol]]
        else:
            return _element_shannon_radii_data_extendedML[symbol]
    else:
        if _print_warnings:
            print("WARNING: Extended Shannon-radius data for element {0} not "
                  "found.".format(symbol))

        return None

# Loader and cache for the element solid-state energy (SSE) datasets.

_element_ssedata = None


def lookup_element_sse_data(symbol):
    """
    Retrieve the solid-state energy (SSE) data for an element.

    Taken from J. Am. Chem. Soc., 2011, 133 (42), pp 16852-16960,
    DOI: 10.1021/ja204670s

    Args:
        symbol : the atomic symbol of the element to look up.

    Returns:
        list : SSE datasets for the element, or None
            if the element was not found among the external data.

        SSE datasets are dictionaries with the keys:

        AtomicNumber
            *int*
        SolidStateEnergy
            *float* SSE
        IonisationPotential
            *float*
        ElectronAffinity
            *float*
        MullikenElectronegativity
            *str*
        SolidStateRenormalisationEnergy
            *float*
    """

    global _element_ssedata

    if _element_ssedata is None:
        _element_ssedata = {}

        with open(os.path.join(data_directory,
                               "SSE.csv"), 'r') as file:
            reader = csv.reader(file)

            for row in reader:
                dataset = {
                    'AtomicNumber': int(row[1]),
                    'SolidStateEnergy': float(row[2]),
                    'IonisationPotential': float(row[3]),
                    'ElectronAffinity': float(row[4]),
                    'MullikenElectronegativity': float(row[5]),
                    'SolidStateRenormalisationEnergy': float(row[6])
                    }

                _element_ssedata[row[0]] = dataset

    if symbol in _element_ssedata:
        return _element_ssedata[symbol]
    else:
        if _print_warnings:
            print("WARNING: Solid-state energy data for element {0} not"
                  " found.".format(symbol))

        return None

# Loader and cache for the revised (2015) element solid-state energy
# (SSE) datasets.

_element_sse2015_data = None


def lookup_element_sse2015_data(symbol, copy=True):
    """
    Retrieve SSE (2015) data for element in oxidation state.

    Retrieve the solid-state energy (SSE2015) data for an element in an
    oxidation state.  Taken from J. Solid State Chem., 2015, 231,
    pp138-144, DOI: 10.1016/j.jssc.2015.07.037.

    Args:
        symbol : the atomic symbol of the element to look up.
        copy: if True (default), return a copy of the data dictionary,
        rather than a reference to a cached object -- only use
        copy=False in performance-sensitive code and where you are
        certain the dictionary will not be modified!

    Returns:
        list : SSE datasets for the element, or None
            if the element was not found among the external data.

        SSE datasets are dictionaries with the keys:

        OxidationState
            *int*
        SolidStateEnergy2015
            *float* SSE2015
    """

    global _element_sse2015_data

    if _element_sse2015_data is None:
        _element_sse2015_data = {}

        with open(os.path.join(data_directory, "SSE_2015.csv"),
                  'r') as file:
            reader = csv.reader(file)

            for row in reader:
                # Elements can have multiple SSE values depending on
                # their oxidation state

                key = row[0]

                dataset = {
                    'OxidationState': int(row[1]),
                    'SolidStateEnergy2015': float(row[2])}

                if key in _element_sse2015_data:
                    _element_sse2015_data[key].append(dataset)
                else:
                    _element_sse2015_data[key] = [dataset]

    if symbol in _element_sse2015_data:
        if copy:
            return [item.copy() for item in
                    _element_sse2015_data[symbol]]
        else:
            return _element_sse2015_data[symbol]
    else:
        if _print_warnings:
            print("WARNING: Solid-state energy (revised 2015) data for "
                  "element {0} not found.".format(symbol))

        return None

# Loader and cache for the element solid-state energy (SSE) from Pauling
# electronegativity datasets.

_element_ssepauling_data = None


def lookup_element_sse_pauling_data(symbol):
    """Retrieve Pauling SSE data

    Retrieve the solid-state energy (SSEPauling) data for an element
    from the regression fit when SSE2015 is plotted against Pauling
    electronegativity.  Taken from J. Solid State Chem., 2015, 231,
    pp138-144, DOI: 10.1016/j.jssc.2015.07.037

    Args:
    symbol (str) : the atomic symbol of the element to look up.

    Returns: A dictionary containing the SSE2015 dataset for the
        element, or None if the element was not found among the external
        data.
    """

    global _element_ssepauling_data

    if _element_ssepauling_data is None:
        _element_ssepauling_data = {}

        with open(os.path.join(data_directory, "SSE_Pauling.csv"),
                  'r') as file:
            reader = csv.reader(file)

            for row in reader:
                dataset = {
                    'SolidStateEnergyPauling': float(row[1])}

                _element_ssepauling_data[row[0]] = dataset

    if symbol in _element_ssepauling_data:
        return _element_ssepauling_data[symbol]
    else:
        if _print_warnings:
            print("WARNING: Solid-state energy data from Pauling "
                  " electronegativity regression fit for "
                  " element {0} not found.".format(symbol))

        return None



#!/usr/bin/env python
"""
This module can be used to calculate roughly the lattice parameters of a
lattice type, based on the radii of the species on each site.
"""
import numpy as np

def cubic_perovskite(shannon_radius): #Cubic Pervoskite
    '''The lattice parameters of the cubic perovskite structure.

    Args:
    	shannon_radius (list) : The radii of the a,b,c ions

    Returns:
       (tuple):
           float values of lattics constants and
           angles (a, b, c, alpha, beta, gamma)
    '''

    limiting_factors=[2*sum(shannon_radius[1:])]
    a = max(limiting_factors)
    b = a
    c = a
#    space = a * np.sqrt(3) - 2 * shannon_radius[1]
    alpha = 90
    beta = 90
    gamma = 90
    return a,b,c,alpha,beta,gamma

def wurtzite(shannon_radius):
    '''The lattice parameters of the wurtzite structure.

    Args:
    	shannon_radius (list) : The radii of the a,b ions

    Returns:
       (tuple):
           float values of lattics constants and
           angles (a, b, c, alpha, beta, gamma)
    '''
    shannon_radius.sort(reverse=True) # Geometry assumes atom A is larger
    # "Ideal" wurtzite structure
    # c/a = 1.633, u = 0.375
    #
    alpha = 90
    beta = 90
    gamma = 120
    #
    # Scenario A: A atoms are touching
    #   i.e. height is that of two tetrahegons with side length a
    #    = 2 * sqrt(2/3) * a
    if shannon_radius[0] > 0.817*(shannon_radius[0]+shannon_radius[1]):
        a = 2 * shannon_radius[0]
        b = a
        c = 2 * np.sqrt(2./3.) * a
    else:
        # Scenario B: regular wurtzite, similar sizes
        a = 2*0.817*(shannon_radius[0]+shannon_radius[1])  # 0.817 is sin(109.6/2)
        b = a
        c = (shannon_radius[0]+shannon_radius[1])*(2+2*0.335)  # 0.335 is sin(109.6-90)
#    inner_space = a * (6**0.5) - (4*shannon_radius[0])
    return a,b,c,alpha,beta,gamma

#A1#
def fcc(covalent_radius):
    '''The lattice parameters of the A1.

	Args:
    	shannon_radius (list) : The radii of the a ions

    Returns:
       (tuple):
           float values of lattics constants and
           angles (a, b, c, alpha, beta, gamma)
    '''
    a = 2 * 2**0.5 * covalent_radius
    b = 2 * 2**0.5 * covalent_radius
    c = 2 * 2**0.5 * covalent_radius
    alpha = 90
    beta = 90
    gamma = 90
    return a,b,c,alpha,beta,gamma

#A2#
def bcc(covalent_radius):
    '''The lattice parameters of the A2.

    Args:
    	shannon_radius (list) : The radii of the a ions

    Returns:
       (tuple):
           float values of lattics constants and
           angles (a, b, c, alpha, beta, gamma)
    '''
    a = 4 * covalent_radius / np.sqrt(3)
    b = a
    c = a
    alpha = 90
    beta = 90
    gamma = 90
    return a,b,c,alpha,beta,gamma

#A3#
def hcp(covalent_radius):
    '''The lattice parameters of the hcp.

    Args:
    	shannon_radius (list) : The radii of the a ions

    Returns:
       (tuple):
           float values of lattics constants and
           angles (a, b, c, alpha, beta, gamma)
    '''
    a = 2 * covalent_radius
    b = a
    c = (4./3.) * 6**0.5 * covalent_radius
    alpha = 90
    beta = 90
    gamma = 120
    return a,b,c,alpha,beta,gamma

#A4#
def diamond(covalent_radius):
    '''The lattice parameters of the diamond.

    Args:
    	shannon_radius (list) : The radii of the a ions

    Returns:
       (tuple):
           float values of lattics constants and
           angles (a, b, c, alpha, beta, gamma)
    '''
    a = 8 * covalent_radius / np.sqrt(3)
    b = a
    c = a
    alpha = 90
    beta = 90
    gamma = 90
    return a,b,c,alpha,beta,gamma

#A5#
def bct(covalent_radius):
    '''The lattice parameters of the bct.

    Args:
    	shannon_radius (list) : The radii of the a ions

    Returns:
       (tuple):
           float values of lattics constants and
           angles (a, b, c, alpha, beta, gamma)
    '''
    a = 3.86 * covalent_radius
    b = a
    c = 2 * covalent_radius
    alpha = 90
    beta = 90
    gamma = 90
    return a,b,c,alpha,beta,gamma

#B1
def rocksalt(shannon_radius):
    '''The lattice parameters of rocksalt.

    Args:
    	shannon_radius (list) : The radii of the a,b ions

    Returns:
       (tuple):
           float values of lattics constants and
           angles (a, b, c, alpha, beta, gamma)
    '''
    limiting_factors=[2*2**0.2*shannon_radius[0],2*2**0.2*shannon_radius[1],2*shannon_radius[0]+2*shannon_radius[1]]
    a = max(limiting_factors)
    b = a
    c = a
    alpha = 90
    beta = 90
    gamma = 90
    return a,b,c,alpha,beta,gamma

#B2
def b2(shannon_radius):
    '''The lattice parameters of b2.

    Args:
    	shannon_radius (list) : The radii of the a,b ions

    Returns:
       (tuple):
           float values of lattics constants and
           angles (a, b, c, alpha, beta, gamma)
    '''
    limiting_factors=[2*(shannon_radius[0]+shannon_radius[0])/np.sqrt(3),2*shannon_radius[1],2*shannon_radius[0]]
    a = max(limiting_factors)
    b = a
    c = a
    alpha = 90
    beta = 90
    gamma = 90
    return a,b,c,alpha,beta,gamma

#B3
def zincblende(shannon_radius):
    '''The lattice parameters of Zinc Blende.
	Args:
    	shannon_radius (list) : The radii of the a,b ions

    Returns:
       (tuple):
           float values of lattics constants and
           angles (a, b, c, alpha, beta, gamma)
    '''
    limiting_factors=[2*(max(shannon_radius)*np.sqrt(2)), 4*(shannon_radius[0] + shannon_radius[1])**(1./3.)]
    a = max(limiting_factors)
    b = a
    c = a
    alpha = 90
    beta = 90
    gamma = 90
    return a,b,c,alpha,beta,gamma


## Zn-S-Zn angle is ~109.5 degrees (from a tetrahedron). It is exactly 2*invCos(-1/3).
## The distance of that Zn-Zn (diagonally to half the face) is (using the cosine rule) is
# root[2(r1+r2)^2 - 2(r1+r2)^(2)cos(ZnSZn angle)].

#B10
def b10(shannon_radius): #Litharge
    '''The lattice parameters of Litharge

    Args:
    	shannon_radius (list) : The radii of the a,b ions

    Returns:
       (tuple):
           float values of lattics constants and
           angles (a, b, c, alpha, beta, gamma)
    '''
    limiting_factors=[4*(max(shannon_radius))/np.sqrt(2), sum(shannon_radius)*1.31]## Explained below.
    a = max(limiting_factors)
    b = a
    c = a*1.26 #Value taken for PbO http://www.mindat.org/min-2466.html#
    alpha = 90
    beta = 90
    gamma = 90
    return a,b,c,alpha,beta,gamma

def stuffed_wurtzite(shannon_radii):
    ''' The stuffed wurtzite structure (e.g. LiGaGe) space group P63/mc.

    Args:
    	shannon_radius (list) : The radii of the a,b,c ions

    Returns:
       (tuple):
           float values of lattics constants and
           angles (a, b, c, alpha, beta, gamma)
    '''
    rac = shannon_radii[2]+ shannon_radii[1]
    x = rac*np.sin(np.radians(19.5))
    c = 2*rac + x
    y = rac*np.sin(np.radians(70.5))
    a = y*np.sin(np.radians(120))/np.sin(np.radians(30))
    b = a
    alpha = 90
    beta  = 90
    gamma = 120
    return a,b,c,alpha,beta,gamma



#!/usr/bin/env python

__author__ = "Daniel W. Davies"
__copyright__ = "Copyright Daniel W. Davies, Adam J. Jackson, Keith T. Butler (2019)"
__version__ = "2.3.2"
__maintainer__ = "Anthony O. Onwuli"
__email__ = "anthony.onwuli16@imperial.ac.uk"
__date__ = "Nov 4 2021"

from setuptools import setup, Extension
import os
import unittest

module_dir = os.path.dirname(os.path.abspath(__file__))

if __name__ == "__main__":
    setup(
        name="SMACT",
        version="2.3.2",
        description="Semiconducting Materials by Analogy and Chemical Theory",
        long_description=open(os.path.join(module_dir, "README.md")).read(),
        long_description_content_type="text/markdown",
        url="https://github.com/WMD-group/SMACT",
        author="Daniel W. Davies",
        author_email="d.davies16@imperial.ac.uk",
        maintainer="Anthony O. Onwuli",
        maintainer_email="anthony.onwuli16@imperial.ac.uk",
        license="MIT",
        packages=["smact", "smact.tests", "smact.structure_prediction"],
        package_data={
            "smact": [
                "data/*.txt",
                "data/*.csv",
                "data/*.data",
                "data/*.xlsx",
                "data/*.json",
            ]
        },
        zip_safe=False,
        test_suite="smact.tests.test",
        install_requires=[
            "scipy",
            "numpy",
            "spglib",
            "pymatgen",
            "ase",
            "pandas",
            "pathos",
        ],
        classifiers=[
            "Programming Language :: Python",
            "Development Status :: 5 - Production/Stable",
            "Intended Audience :: Science/Research",
            "Operating System :: OS Independent",
            "Topic :: Scientific/Engineering",
        ],
    )
